/*
    Copyright (C) 2021  HisArmMan

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <goodbrew/config.h>
#include <goodbrew/base.h>
#include <goodbrew/graphics.h>
#include <goodbrew/images.h>
#include <goodbrew/controls.h>
#include <goodbrew/useful.h>
#include <goodbrew/sound.h>
#include <goodbrew/text.h>
#include <goodbrew/paths.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "goodLinkedList.h"
#include "gamemisc.h"
#include "assistance.h"
#include "main.h"
#include "keyboard.h"

//#define BOARDIMAGEMAKER

#define VERSIONSTRING "v1.0"

//#define DEBUGOUT
//#define CURSORTEST
//#define NOGENINITIALBOARD
#define SQUISHYCURSOR
#define SCOREDIGITS "08"

#define POPUPTIMERATIO 1.2

// for block pop animation. these are timestamps relative to the total time
#define HEARTPULSEUPTIME .2
	#define HEARTPULSEEXTRASCALE .5
#define HEARTDOWNTIME (.4+HEARTPULSEUPTIME)

#define DEFAULTPLAYERPLACETIME 200
#define DEFAULTUPSHIFTANIMTIME 100
#define DEFAULTCURSORORIENT 1 // if you change this, also change the next box preview
#define CURSORROTATETIME 100

#define FULLROTRAD ((M_PI*2))
#define ONEFOURTHRAD ((M_PI/2))
#define HEARTSPINTIME 1000

// an integer representing rotation in increments of PI/2 radians
typedef signed char numorient;

struct drawableblock{
	crossTexture* top;
	crossTexture* mid;
	crossTexture* bottom;
	crossTexture* heart;
};

enum blockcolor{
	COLOR_NONE=0, // reserved 0.
	COLOR_GARBAGE, // 
	COLOR_REAL, // all colors >= this one are real blocks. must not be 0.
};

// whatever info is required to draw the block. may also be needed for processing
struct myblockshape{
	enum blockcolor c;
	// blocks that are placed down are considered upright or sideways
	// and they always extend top down and left to right
	short len;
	// number of parts of the block not on the board yet
	// if not 0, this block is not fully on the board yet and the "len" variable is too short.
	// aside from drawing, this variable can be ignored and the block treated as normal with length "len"
	// if nonzero, block is implicitly vertical.
	char partialleft;
	
	short numhearts;
	short* heartloc;
	u64 spawntime; // used to calculate animation for star and such
};
struct myplacedblock{
	struct myblockshape s;
	
	unsigned char isvertical;
	short x;
	short y;
	
	// if nonzero, that means that the Y position is 1 lower and we're falling from the tile above it
	// after we reach this time, check if we need to fall another tile
	// also, don't do any star touching checks if this number isn't 0.
	u64 endFallTime;
	mltime deltaFallTime; // if i ever get crunched on memory, move this, the color, and the star data all into a "blocktype" struct which i store a pointer to.
	short curchainvalue; // if it connects with another block after this current fall, this determines the chain length of that connection. only valid when block is falling.
	short ifstartsfallchainvalue; // if this block begins to fall on this frame, use this chain value
	
	u64 startPopTime;
	mltime deltaPopTime;
	short popcomboid; // only use if startPopTime==fTime

	u64 endPlayerPlaceTime;
};
struct chainpopup{
	u64 timeend;
	int x;
	int y;
	int value;
	char ischain;
};
// for (struct boardentry).
// is a bitamp.
enum boardentrytype{
	BOARDENT_BLOCKPART=0x1,
	BOARDENT_STAR=0x2,
	BOARDENT_GARBAGE=0x4,
	BOARDENT_HEAD=0x80, // can be combined with another modifier
};
struct boardentry{
	enum boardentrytype type;
};
struct boardentry_block{
	struct boardentry p;
	struct myplacedblock* head;
};
struct scoreinfo{
	u64 score;
	mltime timepassed;
	int maxcombo;
	int maxchain;
	int blockscleared;
};
struct gboard{
	struct nList* blocklist; // of (struct myplacedblock)
	short w;
	short h;
	struct boardentry*** board;
	char gravityqueued;
	int numnextpieces;
	struct myblockshape** nextpieces;
	struct nList* popups; // of (struct chainpopup)

	// last blocks to be generated to the bottom of the board. these blocks must be upheld by the next generation
	// note that these are just copies of the blocks. free them yourself. the actual blocks that were put on the board could be gone already.
	struct nList* lastbottomgen;
	
	u64 upshiftEndTime;
	mltime deltaUpshift;

	u64 nextboxanimend;
	char nextboxanimpiecesconsumed;
	mltime nextboxanimdelta;

	u64 queuedUpshiftTime;
	u64 diedat;
	
	struct scoreinfo score;
	u64 starttime;
};
////
enum bitmapdir{
	DIRBIT_LEFT=1,
	DIRBIT_RIGHT=2,
	DIRBIT_UP=4,
	DIRBIT_DOWN=8,
};
struct cursordas{
	u64 nextDasShift[4];
	enum bitmapdir dasdirs;
	int delay;
	int repeat;
	int* x;
	int* y;
	int minx;
	int maxx;
	int miny;
	int maxy;
};
struct playercursor{
	int x;
	int y;
	struct myblockshape* block;
	struct gboard* board;
	numorient o;
	struct cursordas das;

	char rotateAnimIsClockwise;
	u64 rotateAnimEnd; // offset by CURSORROTATETIME
	
	int moveAnimOrigX;
	u64 moveAnimXEndTime;
	int moveAnimOrigY;
	u64 moveAnimYEndTime; // offset by CURSORROTATETIME
};
struct gamestate{
	struct playablegame* game;
	u64 fTime;
};
struct playablegame{
	char* highscoredatapath;
	struct scoreinfo highscores;
	
	struct gamesettings* settings;
	struct pcontrols* controls;
	void(*initialfunc)(struct gamestate*);
	void(*drawfunc)(struct gamestate*);
	// 0 - do nothing
	// 1 - play again
	// >=2 - quit and return the ret value
	//
	// the main menu interpretes the >=2 return values like this:
	// 2 - normal
	// 3 - go to title screen
	//
	int(*updatefunc)(struct gamestate*);
	
	struct gamestate*(*newstate)(struct playablegame*);
	void(*freestate)(struct gamestate*);
	void(*freeplayable)(struct playablegame*); // free anything else that is in an extended playablegame struct
	crossMusic* music;
	crossPlayHandle musichandle;
	char musicstarted;
};

enum gsound{
	GSOUND_STARCONNECT=0,
	GSOUND_PLACEBLOCK=1,
	GSOUND_MAX
};
char framegsounds[GSOUND_MAX];
crossSE* gsoundsloaded[GSOUND_MAX];

/* struct gpretty{ */
/* 	void (*drawframe)(struct gboard*, struct playercursor*, u64); */
/* }; */
/* struct gpretty_cardboard{ */
/* 	struct pretty parent; */
/* }; */
//////////////////////
static struct myblockshape* genblockalloc(struct gamesettings* settings, u64 fTime);
static char spaceisfree(int x, int y, struct gboard* board);
static char isoob(struct gboard* b, int x, int y);
static int getcursorcenter(int len);
static void placeplacedblock(struct myplacedblock* b, struct gboard* board, u64 fTime);
static void offsettoactualboardcontentpos(int* origx, int* origy);
static void beginboardupshift(struct gboard* board, struct playercursor* cursor, struct gamesettings* se, u64 fTime);
static void freeplacedblock(struct myplacedblock* b);
static void freeblockshape(struct myblockshape* s);
static int getnextboxw(struct gamesettings* se);
static void _drawlabelat(int middlex, int y, int stampsize, unsigned char stampa, unsigned char texta, const char* text);
//////////////////////
int blockw=30;
char* vitaAppId="HAPPYBLOK";
//
int maxcolors;
static struct drawableblock* drawcolors;
static struct drawableblock* cursorcolors;
static crossTexture* backgroundimg;
struct stampres* stampres;
crossFont* mainfont;
static crossSE* countdownSE[2];
int mainFontHeight;
static struct pcontrols globalcontrols;
crossTexture* handimg;
crossSE* diesound;
crossSE* menuselectsound;
static crossSE* winsound;
static char* scorefilesdir;
static char* customgamescorefilesdir;
static char* custommodedir;
mltime preferreddasdelay;
mltime preferreddasrepeat;
static char isgod=0;
//////////////////////
///////////////////////////////////////////////
// MISC
///////////////////////////////////////////////
#if GBPLAT == GB_VITA
// to get sdl2 to link
void vita2d_draw_texture_part_scale(const vita2d_texture *texture, float x, float y, float tex_x, float tex_y, float tex_w, float tex_h, float x_scale, float y_scale){}
void vita2d_draw_texture_tint_part_scale(const vita2d_texture *texture, float x, float y, float tex_x, float tex_y, float tex_w, float tex_h, float x_scale, float y_scale, unsigned int color){}
#endif
static char* getisgodpath(){
	return fixPathAlloc("isgod",TYPE_DATA);
}
static crossSE* loadputstampse(){
	crossSE* putse = loadvpkse("assets/se/mouseclick.ogg");
	setSFXVolumeBefore(putse, 128/2);
	return putse;
}
static char* getsavedcontrolspath(){
	return fixPathAlloc("controls",TYPE_DATA);
}
char menuselectpressed(){
	return wasJustPressed(BUTTON_A) || wasJustPressed(BUTTON_START);
}
static void cleardebugfile(){
	#if GBPLAT == GB_VITA
	char* logfilepath=fixPathAlloc("log.txt", TYPE_DATA);
	FILE* fp = fopen(logfilepath,"wb");
	if (fp){
		fclose(fp);
	}
	free(logfilepath);
	#else
	puts("Welcome");
	#endif
}
void WriteToDebugFile(const char* _formatString, ...){
	va_list _tempArgs;
	va_start(_tempArgs, _formatString);
	char* _completeString = formatf(_tempArgs,_formatString);
	#if GBPLAT == GB_LINUX
		printf("%s",_completeString);
	#else
		char* _tempDebugFileLocationBuffer=fixPathAlloc("log.txt", TYPE_DATA);
		FILE *fp;
		fp = fopen(_tempDebugFileLocationBuffer, "a");
		if (!fp){
			sadflash();
		}else{
			fwrite(_completeString,1,strlen(_completeString),fp);
			fclose(fp);
		}
		free(_tempDebugFileLocationBuffer);
	#endif
	free(_completeString);
}
static int writescorefile(FILE* fp, struct scoreinfo* s){
	if (fputc(1,fp)!=1){
		return 1;
	}
	return fwrite(s,sizeof(struct scoreinfo),1,fp)!=1;
}
static int loadscorefile(FILE* fp, struct scoreinfo* s){
	if (fgetc(fp)!=1){
		return 1;
	}
	return fread(s,sizeof(struct scoreinfo),1,fp)!=1;
}
double getspeedlv(mltime nextupshift){
	return (MINLEVELUPSHIFTTIME-nextupshift)/(double)(MINLEVELUPSHIFTTIME/100);
}
static void resetgsound(){
	memset(framegsounds,0,GSOUND_MAX);
}
static void queuegsound(enum gsound s){
	framegsounds[s]=1;
}
static void playgsound(){
	for (int i=0;i<GSOUND_MAX;++i){
		if (framegsounds[i]){
			playSound(gsoundsloaded[i], 0);
		}
	}
}
#define CHAINMULTDENOM 10 // where the big spike is
static u64 getscorefor(int chainnum, int combonum, mltime nextupshifttime){
	// calculate base score based on next upshift time
	int basescore=(1/(nextupshifttime/(double)500))*1000; //(1/(x/500))*1000
	double chainmult=(pow(chainnum/(double)CHAINMULTDENOM,3)+chainnum/(double)CHAINMULTDENOM)*8+.2;
	double combomult=pow(2,combonum-6)+1;
	return basescore*chainmult*combomult;
}
float fixX(float x){
	return x;
}
float fixY(float y){
	return y;
}
void XOutFunction(){
	exit(1);
}
static int initgfx(){
	initGraphics(960,544,0);
	#ifndef BOARDIMAGEMAKER
	setClearColor(255,0,0);
	#else
	setClearColor(0xf2,0xf2,0xf2);
	#endif
	setWindowTitle("star game1");
	return 0;
}
static int init(){
	#if GBPLAT == GB_LINUX
	fixPathOverride("./vpkcontents/", TYPE_EMBEDDED);
	fixPathOverride("./HAPPYBLOK/", TYPE_DATA);
	#endif
	srand(time(NULL));
	generalGoodInit();
	controlsInit();
	{
		// init dirs
		makedatadir("");

		scorefilesdir=fixPathAlloc("scores/",TYPE_DATA);
		createDirectory(scorefilesdir);

		customgamescorefilesdir=easyCombineStrings(2,scorefilesdir,"customgame/");
		createDirectory(customgamescorefilesdir);

		custommodedir=fixPathAlloc("customgames/",TYPE_DATA);
		createDirectory(custommodedir);

		char* trophypath=getisgodpath();
		isgod=checkFileExist(trophypath);
		free(trophypath);
	}
	cleardebugfile();
	initgfx();
	backgroundimg=loadvpkpng("assets/img/background.png");
	stampres=loadstampres();
	maxcolors=5;
	drawcolors=malloc(sizeof(struct drawableblock)*maxcolors);
	cursorcolors=malloc(sizeof(struct drawableblock)*maxcolors);
	char path[100];
	for (int i=0;i<maxcolors;++i){
		char numchar='0'+i;
		sprintf(path,"assets/img/blocks/%c/top.png",numchar);
		drawcolors[i].top=loadvpkpng(path);
		sprintf(path,"assets/img/blocks/%c/mid.png",numchar);
		drawcolors[i].mid=loadvpkpng(path);
		sprintf(path,"assets/img/blocks/%c/bottom.png",numchar);
		drawcolors[i].bottom=loadvpkpng(path);
		sprintf(path,"assets/img/blocks/%c/heart.png",numchar);
		drawcolors[i].heart=loadvpkpng(path);
		// cursor
		sprintf(path,"assets/img/blocks/%c/cursor/top.png",numchar);
		cursorcolors[i].top=loadvpkpng(path);
		sprintf(path,"assets/img/blocks/%c/cursor/mid.png",numchar);
		cursorcolors[i].mid=loadvpkpng(path);
		sprintf(path,"assets/img/blocks/%c/cursor/bottom.png",numchar);
		cursorcolors[i].bottom=loadvpkpng(path);
			
		cursorcolors[i].heart=drawcolors[i].heart;
	}
	handimg=loadvpkpng("assets/img/ui/hand.png");

	char* fontpath=fixPathAlloc("assets/ume-tgc5.ttf",TYPE_EMBEDDED);
	mainfont=loadFont(fontpath, getScreenHeight()/20);
	free(fontpath);
	mainFontHeight=textHeight(mainfont);
	
	initAudio();
	gsoundsloaded[GSOUND_STARCONNECT]=loadvpkse("assets/se/completetask_0.ogg");
	gsoundsloaded[GSOUND_PLACEBLOCK]=loadvpkse("assets/se/placeblock.ogg");
	setSFXVolumeBefore(gsoundsloaded[GSOUND_PLACEBLOCK], 128*.7);
	countdownSE[0]=loadvpkse("assets/se/countdown/2.ogg");
	countdownSE[1]=loadvpkse("assets/se/countdown/4.ogg");

	diesound=loadvpkse("assets/se/gameover.ogg");
	setSFXVolumeBefore(diesound, 128/2);
	winsound=loadvpkse("assets/se/winsound1-2.ogg");
	menuselectsound=loadvpkse("assets/se/switch35.ogg");

	char* controlspath=getsavedcontrolspath();
	loadcontrols(controlspath, &globalcontrols, &preferreddasdelay, &preferreddasrepeat);
	free(controlspath);
	return 0;
}
///////////////////////////////////////////////
// CURSOR FUNCTIONS
///////////////////////////////////////////////
static void _cursordasonpush(struct cursordas* d, enum bitmapdir dir, u64 fTime){
	d->dasdirs|=d->dasdirs | dir;
	d->nextDasShift[dir==DIRBIT_DOWN ? 3 : dir/2]=fTime+tohdtime(d->delay);
}
static char processcursordas(struct playercursor* c, struct cursordas* d, struct pcontrols* controls, u64 fTime){
	enum bitmapdir movedirs=0;
	enum bitmapdir helddirs=0;
	if (buttondown(&controls->up)){
		helddirs|=DIRBIT_UP;
		if (buttonpressed(&controls->up)){
			movedirs|=DIRBIT_UP;
			_cursordasonpush(d,DIRBIT_UP,fTime);
		}
	}
	if (buttondown(&controls->left)){
		helddirs|=DIRBIT_LEFT;
		if (buttonpressed(&controls->left)){
			movedirs|=DIRBIT_LEFT;
			_cursordasonpush(d,DIRBIT_LEFT,fTime);
		}
	}
	if (buttondown(&controls->right)){
		helddirs|=DIRBIT_RIGHT;
		if (buttonpressed(&controls->right)){
			movedirs|=DIRBIT_RIGHT;
			_cursordasonpush(d,DIRBIT_RIGHT,fTime);
		}
	}
	if (buttondown(&controls->down)){
		helddirs|=DIRBIT_DOWN;
		if (buttonpressed(&controls->down)){
			movedirs|=DIRBIT_DOWN;
			_cursordasonpush(d,DIRBIT_DOWN,fTime);
		}
	}
	if (helddirs & d->dasdirs){
		for (int i=0;i<4;++i){
			enum bitmapdir thisdir=1<<i;
			if ((d->dasdirs & thisdir)){
				if (helddirs & thisdir){
					if (fTime>=d->nextDasShift[i]){
						movedirs|=thisdir;
						d->nextDasShift[i]=fTime+tohdtime(d->repeat);
					}
				}else{
					(d->dasdirs)^=thisdir;
				}
			}
		}
	}
	if (c->moveAnimYEndTime){
		if ((c->moveAnimOrigY<d->miny && helddirs & DIRBIT_UP)
			|| (c->moveAnimOrigY>d->maxy && helddirs & DIRBIT_DOWN)
			){
			c->moveAnimYEndTime=fTime+tohdtime(CURSORROTATETIME/2);
		}
	}
	if (c->moveAnimXEndTime){
		if ((c->moveAnimOrigX<d->minx && helddirs & DIRBIT_LEFT)
			|| (c->moveAnimOrigX>d->maxx && helddirs & DIRBIT_RIGHT)
			){
			c->moveAnimXEndTime=fTime+tohdtime(CURSORROTATETIME/2);
		}
	}
	if (movedirs){
		if (movedirs & DIRBIT_UP){
			--*d->y;
		}
		if (movedirs & DIRBIT_DOWN){
			++*d->y;
		}
		if (movedirs & DIRBIT_LEFT){
			--*d->x;
		}
		if (movedirs & DIRBIT_RIGHT){
			++*d->x;
		}
		#ifndef SQUISHYCURSOR
		limitNumW(d->y,d->miny,d->maxy);
		limitNumW(d->x,d->minx,d->maxx);
		#endif
		return 1;
	}
	return 0;
}
static void startcursormovetowithtime(struct playercursor* cursor, int newx, int newy, mltime movetime, u64 fTime){
	u64 endtime=fTime+tohdtime(movetime);
	if (cursor->x!=newx){
		cursor->moveAnimXEndTime=endtime;
		cursor->moveAnimOrigX=cursor->x;
		cursor->x=newx;
	}
	if (cursor->y!=newy){
		cursor->moveAnimYEndTime=endtime;
		cursor->moveAnimOrigY=cursor->y;
		cursor->y=newy;
	}
}
static void startcursormoveto(struct playercursor* cursor, int newx, int newy, u64 fTime){
	startcursormovetowithtime(cursor, newx, newy, CURSORROTATETIME, fTime);
}
// assumng retx and rety already have the cursor's rotation point in them, offset them so they are at the cursor's top left
static void offsetforcursortopleft(int blocklen, numorient o, int* retx, int* rety){
	int longextra=getcursorcenter(blocklen);
	int shortextra=blocklen-longextra-1;
	if (o==3){
		*retx-=longextra;
	}else if (o==2){
		*rety-=shortextra;
	}else if (o==1){
		*retx-=shortextra;
	}else if (o==0){
		*rety-=longextra;
	}
}
static signed char cursoroverair(struct playercursor* c, int yoffset, struct gboard* board){
	int cx=c->x;
	int cy=c->y+yoffset;
	offsetforcursortopleft(c->block->len, c->o, &cx, &cy);
	int* offsetme=((c->o==0 || c->o==2) ? &cy : &cx); // which direction to offset to move forward
	for (int i=0;i<c->block->len;++i){
		if (isoob(board,cx,cy) || board->board[cx][cy]){
			return 0;
		}
		++*offsetme;
	}
	return 1;
}
// x and y are board positions of the oriented block
// do not free the array inside of the block shape. it is stolen.
static void orientedtoboard(struct myblockshape* s, int x, int y, numorient o, struct myplacedblock* retinfo){
	memcpy(&(retinfo->s),s,sizeof(struct myblockshape));

	if (o==1 || o==2){
		int offset=(s->len-1);
		if (retinfo->s.numhearts==1){
			retinfo->s.heartloc[0]=offset-retinfo->s.heartloc[0];
		}else{
			for (int i=0;i<retinfo->s.numhearts/2;++i){
				// make the star distances relative to the other end.
				// because the array needs to stay sorted, reverse the array after deciding the new distance
				int otherindex=retinfo->s.numhearts-i-1;
				int newfirstloc=offset-retinfo->s.heartloc[0];
				retinfo->s.heartloc[0]=offset-retinfo->s.heartloc[otherindex];
				retinfo->s.heartloc[otherindex]=newfirstloc;
			}
		}
		retinfo->isvertical=(o==2);
	}else{
		retinfo->isvertical=(o==0);
	}
	offsetforcursortopleft(retinfo->s.len,o,&x,&y);
	retinfo->x=x;
	retinfo->y=y;
}
static void fixcursor(struct playercursor* cursor, numorient cursororientation, struct myblockshape* shape, struct gboard* board, mltime cursormovetime, u64 fTime){
	struct cursordas* das = &cursor->das;
	int longextra=getcursorcenter(shape->len);
	int shortextra=shape->len-longextra-1;
	das->minx=0;
	das->miny=0;
	das->maxx=board->w-1;
	das->maxy=board->h-1;
	if (cursororientation==0){
		das->miny+=longextra;
		das->maxy-=shortextra;
	}else if (cursororientation==1){
		das->minx+=shortextra;
		das->maxx-=longextra;
	}else if (cursororientation==2){
		das->miny+=shortextra;
		das->maxy-=longextra;
	}else if (cursororientation==3){
		das->minx+=longextra;
		das->maxx-=shortextra;
	}
	// check if we just rotated out of bounds. if so, do the fix animation.
	int fixedx=cursor->x;
	int fixedy=cursor->y;
	limitNumW(&fixedx,das->minx,das->maxx);
	limitNumW(&fixedy,das->miny,das->maxy);
	startcursormovetowithtime(cursor, fixedx, fixedy, cursormovetime, fTime);
}
static void initplacedblock(struct myplacedblock* b){
	b->endFallTime=0;
	b->startPopTime=0;
	b->curchainvalue=999; // easy bug detection
	b->ifstartsfallchainvalue=1;
	b->popcomboid=-2;
	b->endPlayerPlaceTime=0;
}
static void freeplayercursor(struct playercursor* ret){
	freeblockshape(ret->block);
	free(ret->block);
}
static void newplayercursor(struct playercursor* ret, struct gamesettings* se, struct gboard* board, struct myblockshape* initialshapeallocd, u64 fTime){
	ret->block=initialshapeallocd;
	
	ret->board=board;
	ret->x=ezcenter(ret->block->len,board->w)+ret->block->len/2;
	ret->y=ezcenter(1,board->h);
	ret->o=DEFAULTCURSORORIENT;

	ret->moveAnimYEndTime=0;
	ret->moveAnimXEndTime=0;
	ret->rotateAnimEnd=0;
	
	ret->das.dasdirs=0;
	ret->das.delay=preferreddasdelay;
	ret->das.repeat=preferreddasrepeat;
	ret->das.x=&(ret->x);
	ret->das.y=&(ret->y);
	fixcursor(ret, ret->o, ret->block, ret->board, CURSORROTATETIME, fTime);
}
///////////////////////////////////////////////
// BLOCK OP
///////////////////////////////////////////////
static char hasstaratoff(struct myblockshape* s, int off){
	for (int i=0;i<s->numhearts;++i){
		if (s->heartloc[i]>=off){
			if (s->heartloc[i]==off){
				return 1;
			}
			return 0;
		}
	}
	return 0;
}
static void blockdirectionoffsets(char isvertical, signed char* offx, signed char* offy){
	if (isvertical){
		*offx=0;
		*offy=1;
	}else{
		*offx=1;
		*offy=0;
	}
}
static char canstarcollide(struct myplacedblock* b, u64 fTime){
	return !b->endFallTime && (!b->startPopTime  || b->startPopTime==fTime) && !b->endPlayerPlaceTime;
}
static void startblockpop(struct myplacedblock* b, struct gamesettings* se, u64 fTime){
	if (b->startPopTime){
		if (b->startPopTime!=fTime){
			fprintf(stderr,"programming error: tried to start popping a block that's already popping and didn't start popping on thsi frame\n");
		}
		return;
	}
	b->startPopTime=fTime;
	b->deltaPopTime=se->poptime;
}
// check if there's a star at (boardx,boardy) that (struct myplacedblock* b) could collide with if it (b) were next it
// if so, return the block that the other star is on
static struct myplacedblock* hasstarcollideat(struct myplacedblock* b, int boardx, int boardy, struct gboard* board, u64 fTime){
	if (!isoob(board,boardx,boardy) && board->board[boardx][boardy] && (board->board[boardx][boardy]->type & BOARDENT_STAR)){
		struct myplacedblock* otherblock=((struct boardentry_block*)(board->board[boardx][boardy]))->head;
		if (otherblock!=b && otherblock->s.c==b->s.c && canstarcollide(otherblock,fTime)){
			return otherblock;
		}
	}
	return NULL;
}
static void __synclisttocomboid(struct nList* placedblocklist, int comboid){
	ITERATENLIST(placedblocklist,{
			((struct myplacedblock*)_curnList->data)->popcomboid=comboid;
		});
}
// used with active block processing
// if there's a combo, returns a list of the relevant blocks
// to figure out if nextfreecomboid was used, simply check the combo id of the first thing in the list
static struct nList* processblockstarcollision(struct myplacedblock* b, struct nList* globalcombolist, struct gboard* board, struct gamesettings* se, u64 fTime){
	if (!canstarcollide(b,fTime)){
		return 0;
	}
	int thiscomboid=-1;
	struct nList* curcombolist=NULL; // of struct myplacedblock*
	struct nList** combolistadder=initSpeedyAddnList(&curcombolist);
	signed char offx,offy;
	blockdirectionoffsets(b->isvertical, &offx, &offy);
	for (int i=0;i<b->s.numhearts;++i){
		for (int k=0;k<2;++k){
			int checkx=b->x+offx*b->s.heartloc[i];
			int checky=b->y+offy*b->s.heartloc[i];
			int* changeme=(k==0 ? &checkx : &checky);
			for (int j=-1;j<=2;j+=3){
				*changeme+=j;
				// this code will run with the positions next to all four sides of the block
				struct myplacedblock* otherblock=hasstarcollideat(b, checkx, checky, board, fTime);
				if (otherblock){
					if (otherblock->startPopTime==fTime){
						if (otherblock->popcomboid!=thiscomboid){
							// combine the lists
							// set combo id to prevent adding a duplicate of this list
							thiscomboid=otherblock->popcomboid;
							// finish this list and then add it to the end of the other list
							endSpeedyAddnList(combolistadder);
							__synclisttocomboid(curcombolist, thiscomboid);
							struct nList* listentrywithotherlist = getnList(globalcombolist, otherblock->popcomboid);
							appendnList(listentrywithotherlist->data, curcombolist);
							// and now we pretend we were using the other list from the start
							curcombolist=listentrywithotherlist->data;
							combolistadder=initSpeedyAddnList(&curcombolist);
						}
					}else{
						combolistadder=speedyAddnList(combolistadder, otherblock);
						startblockpop(otherblock, se, fTime);
						otherblock->popcomboid=thiscomboid;
						otherblock->curchainvalue=1; // you only get to keep your chain value if you're a block that just hit the ground
					}
				}
			}
		}
	}
	if (!curcombolist){
		return NULL;
	}
	endSpeedyAddnList(combolistadder);
	// add this block, of course.
	startblockpop(b, se, fTime);
	prependnList(&curcombolist)->data=b;
	// new combo
	if (thiscomboid==-1){
		// it'll be added to the end of this list after the function returns, so help yourself to the id
		thiscomboid=nListLen(globalcombolist);
		__synclisttocomboid(curcombolist, thiscomboid);
		return curcombolist;
	}
	return NULL; // dont need to return it because it's an extension to an existing list
}
// used with random block generation
static char blockwouldstarcollide(struct myplacedblock* b, struct gboard* board, u64 fTime){
	signed char offx,offy;
	blockdirectionoffsets(b->isvertical, &offx, &offy);
	for (int i=0;i<b->s.numhearts;++i){
		int thisx=b->x+offx*b->s.heartloc[i];
		int thisy=b->y+offy*b->s.heartloc[i];
		if (hasstarcollideat(b, thisx+1, thisy, board, fTime) ||
			hasstarcollideat(b, thisx-1, thisy, board, fTime) ||
			hasstarcollideat(b, thisx, thisy+1, board, fTime) ||
			hasstarcollideat(b, thisx, thisy-1, board, fTime)){
			return 1;
		}
	}
	return 0;
}
static void freeblockshape(struct myblockshape* s){
	free(s->heartloc);
}
static void freeplacedblock(struct myplacedblock* b){
	freeblockshape(&b->s);
}
static char underblockfree(struct myplacedblock* b, struct gboard* board){
	if (b->isvertical){
		return spaceisfree(b->x,b->y+b->s.len,board);
	}else{
		for (int i=0;i<b->s.len;++i){
			if (!spaceisfree(b->x+i,b->y+1,board)){
				return 0;
			}
		}
		return 1;
	}
}
// the only thing that stops a block from falling is a gravity check.
// returns 0 for blocks that are stable and can't start falling
// returns 1 if the block just started falling
// returns 2 for blocks that continue to fall. the block is still in a tile fall animation, or it fell on air.
// returns 3 if the block was falling before, and is now not. we should do a collision check with this block.
static char blockaddgrav(struct myplacedblock* b, struct gboard* board, struct gamesettings* se, u64 fTime){
	if (b->startPopTime || b->endPlayerPlaceTime){
		return 0;
	}
	if (b->endFallTime && fTime<b->endFallTime){
		return 2;
	}
	char canfall=underblockfree(b, board);
	if (canfall){
		{
			// fix positions in board array
			signed char offx;
			signed char offy;
			blockdirectionoffsets(b->isvertical, &offx,&offy);
			for (int i=b->s.len-1;i>=0;--i){
				board->board[b->x+offx*i][b->y+offy*i+1]=board->board[b->x+offx*i][b->y+offy*i];
				board->board[b->x+offx*i][b->y+offy*i]=NULL;
			}
			board->board[b->x][b->y]=NULL;
		}
		// now move the actual head object
		++b->y;
		b->deltaFallTime=se->falltime;
		char ret=(b->endFallTime ? 2 : 1);
		u64 alreadyserved=(b->endFallTime ? fTime-b->endFallTime : 0);
		b->endFallTime=fTime+tohdtime(b->deltaFallTime)-alreadyserved;
		return ret;
	}else{
		if (b->endFallTime/* && fTime>b->endFallTime*/){
			// if we got here, we already know it's a block that's at the end of its fall animation
			b->endFallTime=0;
			return 3;
		}
	}
	return 0;
}
// do not free the blockshape's internal array
static struct myplacedblock* placecenteredblockat(struct myblockshape* shape, int x, int y, numorient orientation, struct gboard* board, u64 fTime){
	struct myplacedblock* b=malloc(sizeof(struct myplacedblock));
	initplacedblock(b);
	orientedtoboard(shape, x, y, orientation, b);
	placeplacedblock(b, board, fTime);
	return b;
}
///////////////////////////////////////////////
// BOARD
///////////////////////////////////////////////
static struct boardentry* boardentryat(struct gboard* board, int x, int y){
	if (!isoob(board,x,y)){
		return board->board[x][y];
	}
	return NULL;
}
static void __giveboardentrymaybechainvalue(struct boardentry* be, int chainvalue){
	if (be && (be->type & BOARDENT_BLOCKPART)){
		struct myplacedblock* b = ((struct boardentry_block*)be)->head;
		b->ifstartsfallchainvalue=intmax(chainvalue,b->ifstartsfallchainvalue);
	}
}
static void distributemychainvalue(int chainvalue, int extrayoff, struct myplacedblock* b, struct gboard* board, u64 fTime){
	int y=b->y+extrayoff;
	if (b->isvertical){
		__giveboardentrymaybechainvalue(boardentryat(board, b->x, y),chainvalue);
	}else{
		for (int i=0;i<b->s.len;++i){
			__giveboardentrymaybechainvalue(boardentryat(board, b->x+i, y),chainvalue);
		}
	}
}
static double getboardblockoff(struct gboard* board, u64 fTime){
	if (!board->upshiftEndTime){
		return 0;
	}
	return partMoveEmpties(fTime, board->upshiftEndTime, board->deltaUpshift, blockw);
}
static struct myblockshape* boardgivenextpiece(struct gboard* board, struct gamesettings* settings, u64 fTime){
	struct myblockshape* retshape = board->nextpieces[0];
	for (int i=0;i<board->numnextpieces-1;++i){
		board->nextpieces[i]=board->nextpieces[i+1];
	}
	board->nextpieces[board->numnextpieces-1]=genblockalloc(settings, fTime);
	return retshape;
}
static mltime getnextupshifttime(struct gamesettings* s, mltime timepassed){
	// next = e^(mlpassed*-1*delayshortenratio/delayshortentime)*initial
	mltime ret=exp((timepassed*-1*s->delayshortenratio)/(double)s->delayshortentime)*s->initialupshifttime;
	if (ret<=10){
		ret=10;
	}
	return ret;
}
static void initgamesettings(struct gamesettings* s){
	s->numcolors=intmin(3,maxcolors);
	s->falltime=20;
	s->poptime=300;
	s->upshiftanimtime=DEFAULTUPSHIFTANIMTIME;
	s->minimumupshiftblocksgenerated=2;
	s->playerplacetime=DEFAULTPLAYERPLACETIME;
	s->timetodie=3000;

	s->defaultboardw=8;
	s->defaultboardh=17;
	s->minstars=1;
	s->maxstars=2;
	#ifdef CURSORTEST
	s->minblocklen=4;
	s->maxblocklen=4;
	#else
	s->minblocklen=3;
	s->maxblocklen=3;
	#endif
	s->staraddrarity=.2;
	s->stargen=STARGEN_EDGES;

	s->initialupshifttime=3000;
	s->delayshortenratio=.2;
	s->delayshortentime=60000;

	s->doautoupshift=1;
	s->canmanualupshift=1;
	s->isracemode=0;
	s->initialblockgen=10;

	// internal use for specific modes
	s->gengarbage=0;
}
static void writesettings(struct gamesettings* se, FILE* fp){
	fputc(1,fp);
	WRITETHING(se->numcolors,fp);
	WRITETHING(se->falltime,fp);
	WRITETHING(se->poptime,fp);
	WRITETHING(se->upshiftanimtime,fp);
	WRITETHING(se->minimumupshiftblocksgenerated,fp);
	WRITETHING(se->playerplacetime,fp);
	WRITETHING(se->timetodie,fp);

	WRITETHING(se->initialupshifttime,fp);
	WRITETHING(se->delayshortenratio,fp);
	WRITETHING(se->delayshortentime,fp);

	WRITETHING(se->defaultboardw,fp);
	WRITETHING(se->defaultboardh,fp);
	
	WRITETHING(se->minstars,fp);
	WRITETHING(se->maxstars,fp);
	WRITETHING(se->stargen,fp);
	WRITETHING(se->staraddrarity,fp);
	
	WRITETHING(se->minblocklen,fp);
	WRITETHING(se->maxblocklen,fp);

	WRITETHING(se->doautoupshift,fp);
	WRITETHING(se->canmanualupshift,fp);
	WRITETHING(se->isracemode,fp);
	WRITETHING(se->initialblockgen,fp);
}
static int loadsettings(const char* path, struct gamesettings* se){
	FILE* fp = fopen(path,"rb");
	if (!fp){
		return 1;
	}
	if (getc(fp)!=1){
		fclose(fp);
		return 1;
	}
	initgamesettings(se);
	READTHING(se->numcolors,fp);
	READTHING(se->falltime,fp);
	READTHING(se->poptime,fp);
	READTHING(se->upshiftanimtime,fp);
	READTHING(se->minimumupshiftblocksgenerated,fp);
	READTHING(se->playerplacetime,fp);
	READTHING(se->timetodie,fp);

	READTHING(se->initialupshifttime,fp);
	READTHING(se->delayshortenratio,fp);
	READTHING(se->delayshortentime,fp);

	READTHING(se->defaultboardw,fp);
	READTHING(se->defaultboardh,fp);
	
	READTHING(se->minstars,fp);
	READTHING(se->maxstars,fp);
	READTHING(se->stargen,fp);
	READTHING(se->staraddrarity,fp);
	
	READTHING(se->minblocklen,fp);
	READTHING(se->maxblocklen,fp);

	READTHING(se->doautoupshift,fp);
	READTHING(se->canmanualupshift,fp);
	READTHING(se->isracemode,fp);
	READTHING(se->initialblockgen,fp);
	
	fclose(fp);
	return 0;
}
static void initscoreinfo(struct scoreinfo* s){
	s->score=0;
	s->timepassed=0;
	s->maxcombo=0;
	s->maxchain=0;
	s->blockscleared=0;
}
static void newboard(struct gboard* dest, int w, int h, struct gamesettings* se){
	u64 fTime=getHDTime();
	dest->starttime=0; // means the time is not initialized
	dest->diedat=0;
	dest->blocklist=NULL;
	dest->popups=NULL;
	dest->lastbottomgen=NULL;
	dest->upshiftEndTime=0;
	initscoreinfo(&dest->score);
	dest->queuedUpshiftTime=0;
	dest->nextboxanimend=0;
	dest->numnextpieces=2;
	dest->nextpieces=malloc(sizeof(struct myblockshape*)*dest->numnextpieces);
	for (int i=0;i<dest->numnextpieces;++i){
		dest->nextpieces[i]=genblockalloc(se, fTime);
	}
	dest->w=w;
	dest->h=h;
	dest->board=malloc(sizeof(struct boardentry**)*dest->w);
	for (int i=0;i<dest->w;++i){
		dest->board[i]=malloc(sizeof(struct boardentry*)*dest->h);
		for (int j=0;j<dest->h;++j){
			dest->board[i][j]=NULL;
		}
	}
	dest->gravityqueued=0;
}
static void _freeplacedblocklist(struct nList* l){
	ITERATENLIST(l,{
			freeplacedblock(_curnList->data);
		});
	freenList(l,1);
}
static void freeboard(struct gboard* board){
	_freeplacedblocklist(board->blocklist);
	_freeplacedblocklist(board->lastbottomgen);
	for (int i=0;i<board->numnextpieces;++i){
		freeblockshape(board->nextpieces[i]);
		free(board->nextpieces[i]);
	}
	free(board->nextpieces);
	freenList(board->popups,1);
	for (int i=0;i<board->w;++i){
		for (int j=0;j<board->h;++j){
			if (board->board[i][j]){
				// if new board entry types are added, they may need to be freed differently
				free(board->board[i][j]);
			}
		}
		free(board->board[i]);
	}
	free(board->board);
}
static int qsortshortcomparer(const void* a, const void* b){
	return (*((short*)a))-(*((short*)b));
}
static void genblock(struct myblockshape* s, struct gamesettings* settings, u64 fTime){
	s->spawntime=fTime;
	s->c=COLOR_REAL+randbetween(0,settings->numcolors-1);
	s->partialleft=0;
	s->len=randbetween(settings->minblocklen, settings->maxblocklen);
	
	// generate stars
	s->numhearts=settings->minstars;
	for (int i=0;i<settings->maxstars-settings->minstars;++i){
		if (rand()/(double)RAND_MAX<settings->staraddrarity){ // must be strictly less than.
			s->numhearts++;
		}
	}
	if (s->numhearts>s->len){
		s->numhearts=s->len;
	}
	if (s->numhearts>0){
		char dosort=0;
		if (settings->stargen==STARGEN_EDGES || (settings->stargen==STARGEN_EDGESMIDDLE && s->len<=2)){
			if (s->numhearts>2){
				s->numhearts=2;
			}
			s->heartloc=malloc(sizeof(short)*s->numhearts);
			if (s->numhearts==2){
				s->heartloc[0]=0;
				s->heartloc[1]=s->len-1;
			}else{
				s->heartloc[0]=((rand() & 1) ? 0 : s->len-1);
			}
		}else if (settings->stargen==STARGEN_EDGESMIDDLE){
			if (s->numhearts>3){
				s->numhearts=3;
			}
			s->heartloc=malloc(sizeof(short)*s->numhearts);
			int possiblerand[] = {0,s->len/2,s->len-1};
			struct randso* r = newrandspecificonce(possiblerand, 3, -1);
			for (int i=0;i<s->numhearts;++i){
				s->heartloc[i]=randsonext(r);
			}
			freerandso(r);
			dosort=1;
		}else if (settings->stargen==STARGEN_RANDOM){
			s->heartloc=malloc(sizeof(short)*s->numhearts);
			struct randrro* r = newrandrangeonce(0, s->len-1);
			for (int i=0;i<s->numhearts;++i){
				s->heartloc[i]=randrronext(r);
			}
			freerandrro(r);
			dosort=1;
		}
		if (dosort){
			qsort(s->heartloc, s->numhearts, sizeof(short),qsortshortcomparer);
		}
	}else{
		s->heartloc=NULL;
	}
}
static struct myblockshape* genblockalloc(struct gamesettings* settings, u64 fTime){
	struct myblockshape* ret = malloc(sizeof(struct myblockshape));
	genblock(ret, settings, fTime);
	return ret;
}
static void dupshape(struct myblockshape* dest, struct myblockshape* source){
	memcpy(dest,source,sizeof(struct myblockshape));
	dest->heartloc=malloc(sizeof(short)*dest->numhearts);
	memcpy(dest->heartloc,source->heartloc,sizeof(short)*dest->numhearts);
}
static struct myplacedblock* dupplacedblock(struct myplacedblock* source){
	struct myplacedblock* ret = malloc(sizeof(struct myplacedblock));
	memcpy(ret,source,sizeof(struct myplacedblock));
	dupshape(&ret->s, &source->s);
	return ret;
}
static void placeplacedblock(struct myplacedblock* b, struct gboard* board, u64 fTime){
	addnList(&(board->blocklist))->data=b;
	{
		signed char offx,offy;
		blockdirectionoffsets(b->isvertical, &offx, &offy);
		int nextstar=b->s.numhearts>0 ? 0 : -1;
		for (int i=0;i<b->s.len;++i){
			struct boardentry_block* beb = malloc(sizeof(struct boardentry_block));
			beb->p.type=BOARDENT_BLOCKPART;
			beb->head=b;
			if (nextstar!=-1){
				if (b->s.heartloc[nextstar]==i){
					beb->p.type|=BOARDENT_STAR;
					if ((++nextstar)==b->s.numhearts){
						nextstar=-1;
					}
				}
			}
			if (i==0){
				beb->p.type|=BOARDENT_HEAD;
			}
			board->board[b->x+offx*i][b->y+offy*i]=(void*)beb;
		}
	}
}
static char canplaceblock(struct myplacedblock* b, struct gboard* board){
	signed char offx;
	signed char offy;
	blockdirectionoffsets(b->isvertical, &offx, &offy);
	for (int i=0;i<b->s.len;++i){
		if (!spaceisfree(b->x+i*offx, b->y+i*offy, board)){
			return 0;
		}
	}
	return 1;
}
static void generateinitialboard(struct gboard* board, struct gamesettings* se, u64 fTime){
	#ifdef NOGENINITIALBOARD
	return;
	#endif
	for (int i=0;i<se->initialblockgen;++i){
		struct myplacedblock* b=malloc(sizeof(struct myplacedblock));
		initplacedblock(b);
		genblock(&b->s, se, fTime);
		b->isvertical=(rand()&1);
		for (int isverticalloop=0;isverticalloop<2;++isverticalloop){
			b->isvertical=!b->isvertical;
			int maxx;
			if (b->isvertical){
				maxx=board->w-1;
			}else{
				maxx=(board->w-(b->s.len-1))-1;
			}
			struct randrro* xrandomstate = newrandrangeonce(0, maxx);
			while(1){
				b->x=randrronext(xrandomstate);
				if (b->x<0){
					freerandrro(xrandomstate);
					break;
				}
				int i;
				for (i=1;i<board->h;++i){
					b->y=i;
					if (!canplaceblock(b, board)){
						break;
					}
				}
				i--;
				// if we cant place it without killing the player
				if (i==0){
					continue;
				}
				b->y=i;
				// if the block would cause a match, switch its color.
				if (blockwouldstarcollide(b, board, fTime)){
					struct randrro* rstate = newrandrangeonce(COLOR_REAL, COLOR_REAL+se->numcolors-1);
					randrrosetused(rstate, b->s.c);
					int nextcolor;
					while((nextcolor=randrronext(rstate))>=COLOR_REAL){
						b->s.c=nextcolor;
						if (!blockwouldstarcollide(b, board, fTime)){
							break;
						}
					}
					freerandrro(rstate);
					if (nextcolor<COLOR_REAL){
						continue;
					}
				}
				// if we're good:
				freerandrro(xrandomstate);
				goto goodplaceblock;
			}
			if (isverticalloop==1){
				goto failplaceblock;
			}
		}
	goodplaceblock:
		placeplacedblock(b, board, fTime);
		continue;
	failplaceblock:
		freeplacedblock(b);
		free(b);
		continue;
	}
	// find which blocks are at the bottom and register them accordingly
	ITERATENLIST(board->blocklist,{
			struct myplacedblock* b = _curnList->data;
			if (b->y==(b->isvertical ? board->h-b->s.len : board->h-1)){
				addnList(&(board->lastbottomgen))->data=dupplacedblock(b);
			}
		});
}
static char isoob(struct gboard* b, int x, int y){
	return x<0 || y<0 || x>=b->w || y>=b->h;
}
static char spaceisfree(int x, int y, struct gboard* board){
	if (board->upshiftEndTime && y==board->h-1){
		return 0;
	}
	return !isoob(board,x,y) && !(board->board[x][y]);
}
static void boardongamestart(struct gboard* board, struct gamesettings* se, u64 fTime){
	if (se->doautoupshift){
		board->queuedUpshiftTime=fTime+tohdtime(getnextupshifttime(se,board->score.timepassed));
	}else{
		board->queuedUpshiftTime=0;
	}
	board->starttime=fTime;
}
static int boardtoprowfree(struct gboard* board){
	int i;
	for (i=0;i<board->w;++i){
		if (board->board[i][0]){
			return 0;
		}
	}
	return 1;
}
static int onboardqueuedupshift(struct gboard* board, struct playercursor* cursor, u64 upshiftalreadypaid, struct gamesettings* se, u64 fTime){
	// note - if you place a block at the top of the board and then it begins to upshift before the block is out of the playerplace animation, you will begin to die.
	// because the code to fix this would be very complicated, downshifting many blocks, and you don't die instantly, i conclude that this isn't an important problem.
	if (!boardtoprowfree(board)){
		// check if you're dead yet
		u64 dietime=board->queuedUpshiftTime+tohdtime(se->timetodie);
		if (fTime>=dietime){
			char issafe=0;
			ITERATENLIST(board->blocklist,{
					struct myplacedblock* b = _curnList->data;
					if (b->endFallTime || b->startPopTime){
						issafe=1;
						break;
					}
					// if you place a block after the board fills red, you still die.
					if (b->endPlayerPlaceTime && b->endPlayerPlaceTime-tohdtime(se->playerplacetime)<dietime){
						issafe=1;
						break;
					}
				});
			if (!issafe){
				// dead.
				puts("How does it feel to be dead?");
				return 1;
			}
		}
	}else{
		beginboardupshift(board, cursor, se, fTime);
		if (se->doautoupshift){
			// we intentionally ignore queue time already paid because the queue is paused if the top is blocked
			board->queuedUpshiftTime=fTime+tohdtime(getnextupshifttime(se,board->score.timepassed));
		}else{
			board->queuedUpshiftTime=0;
		}
		board->upshiftEndTime-=upshiftalreadypaid;
	}
	return 0;
}
static void updateboard(struct gboard* board, struct playercursor* cursor, struct gamesettings* se, u64 fTime){
	board->score.timepassed=tomltime(fTime-board->starttime);
	if (board->queuedUpshiftTime && fTime>=board->queuedUpshiftTime){
		if (!board->upshiftEndTime){
			if (onboardqueuedupshift(board,cursor,0,se,fTime)){
				board->diedat=fTime;
				return;
			}
		}
	}
	if (board->nextboxanimend && fTime>=board->nextboxanimend){
		board->nextboxanimend=0;
		board->nextboxanimpiecesconsumed--;
		if (board->nextboxanimpiecesconsumed!=0){
			board->nextboxanimend=fTime+tohdtime(board->nextboxanimdelta);
		}
	}
	if (board->upshiftEndTime && fTime>=board->upshiftEndTime){
		board->upshiftEndTime=0;
		board->gravityqueued=1;
	}
	ITERATENLIST(board->popups,{
			struct chainpopup* p = _curnList->data;
			if (fTime>=p->timeend){
				ITERATENLISTREMOVETHISONE(board->popups,1);
			}
		});
	ITERATENLIST(board->blocklist,{
			struct myplacedblock* b = _curnList->data;
			if (b->endFallTime && fTime>=b->endFallTime){
				// the only thing that stops a block from falling is a gravity check.
				// so queue that. for the rest of this frame though, nothing can collide with this block
				board->gravityqueued=1;
			}else if (b->endPlayerPlaceTime && fTime>=b->endPlayerPlaceTime){
				b->endPlayerPlaceTime=0;
				// queue collision check for this block
				board->gravityqueued=1;
				b->endFallTime=10;
				b->deltaFallTime=1;
				b->curchainvalue=1; // you need this if you're going to queue gravity
			}else if (b->startPopTime && fTime>=b->startPopTime+tohdtime(b->deltaPopTime)){
				// gives the blocks above it the chain value
				distributemychainvalue(b->curchainvalue+1, -1, b, board, fTime);
				//
				board->gravityqueued=1;
				// kill block
				// remove from board
				signed char offx;
				signed char offy;
				blockdirectionoffsets(b->isvertical, &offx, &offy);
				for (int i=0;i<b->s.len;++i){
					int thisx=b->x+offx*i;
					int thisy=b->y+offy*i;
					free(board->board[thisx][thisy]); // the board entry doesnt have any malloc inside of it
					board->board[thisx][thisy]=NULL;
				}
				// free and remove actual block from list
				freeplacedblock(_curnList->data);
				ITERATENLISTREMOVETHISONE(board->blocklist,1);
				board->score.blockscleared++;
			}
		});
	if (board->gravityqueued){
		board->gravityqueued=0;
		struct nList* globalcombolist=NULL;
		for (int j=board->h-1;j>=0;--j){
			for (int i=0;i<board->w;++i){
				if (board->board[i][j] && (board->board[i][j]->type & BOARDENT_HEAD)){
					// process it
					struct myplacedblock* b=((struct boardentry_block*)(board->board[i][j]))->head;
					int addgravres=blockaddgrav(b, board, se, fTime);
					if (addgravres==1){
						// new falling block.
						b->curchainvalue=b->ifstartsfallchainvalue;
						distributemychainvalue(b->curchainvalue, -2, b, board, fTime);
					}else if (addgravres==3){
						struct nList* gotcombolist = processblockstarcollision(b, globalcombolist, board, se, fTime);
						if (gotcombolist){
							queuegsound(GSOUND_STARCONNECT);
							addnList(&globalcombolist)->data=gotcombolist;
						}
					}
					b->ifstartsfallchainvalue=1;
				}
			}
		}
		int i=0;
		ITERATENLIST(globalcombolist,{
				struct nList* thiscombolist=_curnList->data;
				if (((struct myplacedblock*)thiscombolist->data)->popcomboid==i){
					int miny=board->h;
					int minx=board->w;
					int totalstars=0;
					int maxchainvalue=0;
					ITERATENLIST(thiscombolist,{
							struct myplacedblock* b = _curnList->data;
							totalstars+=b->s.numhearts;
							maxchainvalue=intmax(b->curchainvalue,maxchainvalue);
							if (b->y<=miny){
								if (b->y<miny || b->x<minx){
									miny=b->y;
									minx=b->x;
								}
							}
						});
					ITERATENLIST(thiscombolist,{
							struct myplacedblock* b = _curnList->data;
							b->curchainvalue=maxchainvalue;
						});

					// chain multiplier:
					u64 newpoints=getscorefor(maxchainvalue,totalstars,getnextupshifttime(se,board->score.timepassed));
					board->score.score+=newpoints;
					printf("%d chain, %d combo. id: %d. get %"PRIu64"\n",maxchainvalue,totalstars,i,newpoints);
					board->score.maxchain=intmax(maxchainvalue,board->score.maxchain);
					board->score.maxcombo=intmax(totalstars,board->score.maxcombo);

					u64 endtime=fTime+tohdtime(se->poptime)*POPUPTIMERATIO;
					if (maxchainvalue>1){
						struct chainpopup* combopop=malloc(sizeof(struct chainpopup));
						combopop->timeend=endtime;
						combopop->x=minx;
						combopop->y=miny;
						combopop->value=maxchainvalue;
						combopop->ischain=1;
						addnList(&board->popups)->data=combopop;
					}
					if (totalstars>2){
						struct chainpopup* combopop=malloc(sizeof(struct chainpopup));
						combopop->timeend=endtime;
						combopop->x=minx;
						combopop->y=miny;
						combopop->value=totalstars;
						combopop->ischain=0;
						addnList(&board->popups)->data=combopop;
					}
				}
				freenList(_curnList->data,0);
				++i;
			});
		freenList(globalcombolist,0);
	}
}
static void playerupdateboard(struct gboard* board, struct playercursor* cursor, struct pcontrols* controls, struct gamesettings* se, u64 fTime){
	if (board->diedat){
		return;
	}
	u64 oldupshifttime=board->upshiftEndTime;
	updateboard(board,cursor,se,fTime);
	if (board->diedat){
		// if we just died
		return;
	}
	if (buttonpressed(&controls->place)){
		// ensure the cursor has the space
		char isok;
		if (!(isok = cursoroverair(cursor, 0, board))){
			int shiftreq;
			// if we're able to place the block by downshifting by 1 or upshifting by 1, then do it.
			if ((isok=cursoroverair(cursor, -1, board))){ // intentional: check upshift first
				shiftreq=-1;
			}else if ((isok=cursoroverair(cursor, 1, board))){
				shiftreq=1;
			}
			if (isok){
				startcursormoveto(cursor, cursor->x, cursor->y+shiftreq, fTime);
			}
		}
		if (isok){
			queuegsound(GSOUND_PLACEBLOCK);
			struct myplacedblock* pb = placecenteredblockat(cursor->block, cursor->x, cursor->y, cursor->o, board, fTime);
			pb->endPlayerPlaceTime=fTime+tohdtime(se->playerplacetime);
			if (board->nextboxanimend){
				board->nextboxanimpiecesconsumed+=1;
				// make the animation shorter
				double curprogress=(board->nextboxanimend-fTime)/(double)board->nextboxanimdelta;
				int newdelta=intmax(board->nextboxanimdelta/2,0);
				board->nextboxanimdelta=newdelta;
				board->nextboxanimend=fTime+(curprogress)*newdelta;
			}else{
				board->nextboxanimpiecesconsumed=1;
				board->nextboxanimend=pb->endPlayerPlaceTime;
				board->nextboxanimdelta=se->playerplacetime;
			}
			free(cursor->block);
			cursor->block=boardgivenextpiece(board, se, fTime);
			cursor->o=DEFAULTCURSORORIENT;
			fixcursor(cursor, cursor->o, cursor->block, cursor->board, CURSORROTATETIME, fTime);
		}
	}
	if (buttonpressed(&controls->cw) || buttonpressed(&controls->cc)){
		cursor->rotateAnimIsClockwise=buttonpressed(&controls->cw);
		cursor->o=wrapnum(cursor->o+(cursor->rotateAnimIsClockwise ? 1 : -1),0,3);
		fixcursor(cursor, cursor->o, cursor->block, board, CURSORROTATETIME, fTime);
		cursor->rotateAnimEnd=fTime+tohdtime(CURSORROTATETIME);
	}
	if (se->canmanualupshift && buttondown(&controls->raisestack) && !(board->upshiftEndTime) && (!board->queuedUpshiftTime || !(fTime>=board->queuedUpshiftTime))){
		board->queuedUpshiftTime=fTime;
		// we just finished an upshift on this frame.
		// take the extra time it had leftover and allocate it to this upshift
		u64 alreadypaid=oldupshifttime ? (fTime-oldupshifttime) : 0;
		onboardqueuedupshift(board, cursor, alreadypaid, se, fTime);
	}
	int oldx=cursor->x;
	int oldy=cursor->y;
	if (processcursordas(cursor,&cursor->das, controls, fTime)){
		fixcursor(cursor, cursor->o, cursor->block, board, CURSORROTATETIME/2, fTime);
		if (cursor->x!=oldx || cursor->y!=oldy){
			//queuegsound(GSOUND_MOVE);
		}
	}
}
///////////////////////////////////////////////
// DRAW
///////////////////////////////////////////////
void drawbackground(){
	#ifndef BOARDIMAGEMAKER
	if (backgroundimg){
		drawTextureSized(backgroundimg,0,0,getScreenWidth(),getScreenHeight());
	}
	#endif
}
static void _drawheartat(int x, int y, enum blockcolor c, u64 spawntime, unsigned char heartalpha, double rotation, float extrascale, u64 fTime){
	c-=COLOR_REAL;
	//int heartoff=blockw/26;
	int heartoff=0;
	rotation+=partMoveFills(fTime, spawntime+tohdtime(HEARTSPINTIME), HEARTSPINTIME, FULLROTRAD);
	int origw=blockw-heartoff*2;
	int heartw=origw*extrascale;
	drawTextureSizedAlphaRotate(drawcolors[c].heart,x+heartoff+ezcenter(heartw,origw),y+heartoff+ezcenter(heartw,origw),heartw,heartw,heartalpha,rotation);
}
static void drawblockpart(int x, int y, struct myblockshape* s, double rotation, u64 startPopTime, mltime deltaPopTime, u64 fTime){
	struct drawableblock* drawable=&drawcolors[s->c-COLOR_REAL];
	int offx=sin(rotation)*blockw*-1;
	int offy=cos(rotation)*blockw;
	unsigned char a=255;
	if (startPopTime){
		u64 enddelta=tohdtime(deltaPopTime);
		u64 curdelta=fTime-startPopTime;
		if (curdelta>=enddelta*HEARTPULSEUPTIME){
			a=partMoveEmpties(fTime, startPopTime+tohdtime(deltaPopTime), deltaPopTime*(1-HEARTPULSEUPTIME), 255);
		}
	}
	int partx=x;
	int party=y;
	for (int i=0;i<s->len;++i){
		crossTexture* tex;
		if (i==0){
			tex=drawable->top;
		}else if (i==s->len-1 && !s->partialleft){
			tex=drawable->bottom;
		}else{
			tex=drawable->mid;
		}
		drawTextureSizedAlphaRotate(tex,partx,party,blockw,blockw,a,rotation);
		partx+=offx;
		party+=offy;
	}
}
// upshiftEndTime, deltaUpshift, distfrombottom (the head), isvertical are all 0 unless drawing as part of the board
static void drawheartpart(int x, int y, struct myblockshape* s, double rotation, u64 startPopTime, mltime deltaPopTime, u64 upshiftEndTime, mltime deltaUpshift, int distfrombottom, char isvertical, u64 fTime){
	int offx=sin(rotation)*blockw*-1;
	int offy=cos(rotation)*blockw;
	float heartscale=1;
	if (startPopTime){
		u64 enddelta=tohdtime(deltaPopTime);
		u64 curdelta=fTime-startPopTime;
		if (curdelta<=enddelta*HEARTPULSEUPTIME){
			mltime pulsedelta=deltaPopTime*HEARTPULSEUPTIME;
			heartscale=1+partMoveFills(fTime, startPopTime+tohdtime(pulsedelta), pulsedelta, HEARTPULSEEXTRASCALE);
		}else if (curdelta<=enddelta*HEARTDOWNTIME){
			mltime downdelta=deltaPopTime*(HEARTDOWNTIME-HEARTPULSEUPTIME);
			heartscale=partMoveEmpties(fTime, startPopTime+tohdtime(deltaPopTime*HEARTDOWNTIME), downdelta, 1+HEARTPULSEEXTRASCALE);
		}else{
			return;
		}
	}
	int numhdraw=s->numhearts;
	if (s->partialleft){ // dont draw unseen hearts
		for (int i=0;i<s->numhearts;++i){
			if (s->heartloc[i]>=s->len){
				numhdraw=i;
				break;
			}
		}
	}
	if (upshiftEndTime && ((!isvertical && distfrombottom==1) || (isvertical && distfrombottom<=s->len))){
		float altscale=partMoveFills(fTime, upshiftEndTime, deltaUpshift, 1);
		if (!isvertical){
			heartscale=altscale;
			goto drawnormal;
		}
		for (int i=0;i<numhdraw;++i){
			_drawheartat(x+offx*s->heartloc[i], y+offy*s->heartloc[i], s->c, s->spawntime, 255, 0, (distfrombottom-s->heartloc[i]==1) ? altscale : heartscale, fTime);
		}
	}else{
	drawnormal:
		for (int i=0;i<numhdraw;++i){
			_drawheartat(x+offx*s->heartloc[i], y+offy*s->heartloc[i], s->c, s->spawntime, 255, 0, heartscale, fTime);
		}
	}
}
static void drawmyblockbothparts(int x, int y, struct myblockshape* s, double rotation, u64 startPopTime, mltime deltaPopTime, u64 fTime){
	drawblockpart(x, y, s, rotation, startPopTime, deltaPopTime, fTime);
	drawheartpart(x, y, s, rotation, startPopTime, deltaPopTime, 0, 0, 0, 0, fTime);
}
static int getcursorcenter(int len){
	return len/2;
}
static void rawdrawcursor(int cx, int cy, struct myblockshape* s, struct drawableblock* drawable, double rotation, u64 fTime){
	#ifdef CURSORTEST
	drawRectangle(cx,cy,blockw,blockw,0,0,0,255);
	#endif
	int centeri=getcursorcenter(s->len);
	int offx=sin(rotation)*blockw*-1;
	int offy=cos(rotation)*blockw;
	int nextstar=(s->numhearts>0 ? 0 : -1);
	for (int i=0;i<s->len;++i){
		int mult=i-centeri;
		int partx=cx+offx*mult;
		int party=cy+offy*mult;
		crossTexture* tex;
		if (i==0){
			tex=drawable->top;
		}else if (i==s->len-1){
			tex=drawable->bottom;
		}else{
			tex=drawable->mid;
		}
		drawTextureSizedAlphaRotate(tex,partx,party,blockw,blockw,255,rotation);
		if (nextstar!=-1){
			if (s->heartloc[nextstar]==i){
				_drawheartat(partx, party, s->c, s->spawntime, 200, 0, 1, fTime);
				if ((++nextstar)==s->numhearts){
					nextstar=-1;
				}
			}
		}
	}
}
static void drawcursor(int boardx, int boardy, struct playercursor* cursor, struct gboard* board,  u64 fTime){
	double rotation=cursor->o*(M_PI/(double)2);
	int drawx=boardx;
	int drawy=boardy;
	offsettoactualboardcontentpos(&drawx,&drawy);
	drawx+=cursor->x*blockw;
	drawy+=cursor->y*blockw;
	if (cursor->moveAnimXEndTime){
		if (fTime<cursor->moveAnimXEndTime){
			int deltax=cursor->moveAnimOrigX-cursor->x;
			drawx+=partMoveEmpties(fTime, cursor->moveAnimXEndTime, CURSORROTATETIME, 1)*deltax*blockw;
		}else{
			cursor->moveAnimXEndTime=0;
		}
	}
	if (cursor->moveAnimYEndTime){
		if (fTime<cursor->moveAnimYEndTime){
			int deltay=cursor->moveAnimOrigY-cursor->y;
			drawy+=partMoveEmpties(fTime, cursor->moveAnimYEndTime, CURSORROTATETIME, 1)*deltay*blockw;
		}else{
			cursor->moveAnimYEndTime=0;
		}
	}
	if (cursor->rotateAnimEnd){
		if (fTime<cursor->rotateAnimEnd){
			rotation+=partMoveEmpties(fTime, cursor->rotateAnimEnd, CURSORROTATETIME, M_PI/2)*(cursor->rotateAnimIsClockwise ? -1 : 1);
		}else{
			cursor->rotateAnimEnd=0;
		}
	}
	drawy+=getboardblockoff(board,fTime);
	rawdrawcursor(drawx,drawy, cursor->block, &(cursorcolors[cursor->block->c-COLOR_REAL]), rotation, fTime);
}
static int __drawboard_bordersize(){
	return ceil(blockw*.5);
}
static int getboardwnonextbox(struct gboard* board){
	return board->w*blockw+__drawboard_bordersize()*2;
}
void getboarddim(struct gboard* board, struct gamesettings* se, int* retw, int* reth){
	int borderoff=__drawboard_bordersize();
	*retw=getboardwnonextbox(board)+getnextboxw(se)-__drawboard_bordersize();
	*reth=board->h*blockw+borderoff*2;
}
void getboarddimbysize(struct gamesettings* se, int boardtilew, int boardtileh, int* retw, int *reth){
	struct gboard board;
	board.w=boardtilew;
	board.h=boardtileh;
	getboarddim(&board,se,retw,reth);
}
static void stdboarddraw(struct gboard* board, struct gamesettings* se, int* retx, int* rety, int* retw, int* reth){
	getboarddim(board,se,retw,reth);
	*rety=ezcenter(*reth,getScreenHeight());
	if (*retw/(double)(getScreenWidth())>=.6){
		if (*rety<blockw/2){
			*retx=*rety;
		}else{
			*retx=blockw/4;
		}
	}else{
		*retx=.08*getScreenWidth();
	}
}
static void _boarddrawrect(int x, int y, int w, int h, unsigned char r, unsigned char g, unsigned char b, unsigned char a, unsigned char redr, unsigned char redg, unsigned char redb, unsigned char reda, int deathminy){
	if (deathminy==-1){
		drawRectangle(x,y,w,h,r,g,b,a);
	}else{
		if (y<deathminy){
			int redh=(y+h)-deathminy;
			if (redh>0){
				drawRectangle(x,y+h-redh,w,redh,redr,redg,redb,reda);
				h-=redh;
			}
			drawRectangle(x,y,w,h,r,g,b,a);
		}else{
			drawRectangle(x,y,w,h,redr,redg,redb,reda);
		}
	}
}
static void boarddrawrect(int x, int y, int w, int h, unsigned char r, unsigned char g, unsigned char b, unsigned char a, int deathminy){
	_boarddrawrect(x, y, w, h, r, g, b, a, 255, 0, 0, 255, deathminy);
}
static void drawboardborder(int x, int y, int wtiles, int htiles, int mindeathy){
	#ifndef BOARDIMAGEMAKER
	int edgew=__drawboard_bordersize();
	boarddrawrect(x, y, wtiles*blockw+edgew*2,htiles*blockw+edgew*2, 0, 0, 0, 255, mindeathy);
	#endif
}
static void offsettoactualboardcontentpos(int* origx, int* origy){
	int borderoff=__drawboard_bordersize();
	*origx+=borderoff;
	*origy+=borderoff;
}
static int getblockfallyoff(struct myplacedblock* b, u64 fTime){
	if (b->endFallTime){
		return -1*partMoveEmpties(fTime, b->endFallTime, b->deltaFallTime, blockw);
	}
	return 0;
}
static void _getblockdrawprops(int boardx, int blocky, int cursorboxx, int cursorboxpiecey, struct myplacedblock* b, float* retx, float* rety, double* retrot, struct gamesettings* se, u64 fTime){
	*retx=boardx+b->x*blockw;
	*rety=blocky+b->y*blockw+getblockfallyoff(b,fTime);
	if (b->endPlayerPlaceTime){
		int diffx=*retx-cursorboxx;
		int diffy=*rety-cursorboxpiecey;
		*retx=cursorboxx+partMoveFills(fTime, b->endPlayerPlaceTime, se->playerplacetime, diffx);
		*rety=cursorboxpiecey+partMoveFills(fTime, b->endPlayerPlaceTime, se->playerplacetime, diffy);
		double defaultrot=ONEFOURTHRAD*3;
		if (b->isvertical){
			*retrot=defaultrot+partMoveFills(fTime, b->endPlayerPlaceTime, se->playerplacetime, ONEFOURTHRAD);
		}else{
			*retrot=defaultrot;
		}
	}else{
		*retrot=b->isvertical ? 0 : ONEFOURTHRAD*3;
	}
}
static void drawnextpieceat(int nextboxx, int nextboxy, struct myblockshape* s, int nextboxw, int nextboxh, u64 fTime){
	drawmyblockbothparts(nextboxx+ezcenter(s->len*blockw,nextboxw)+((s->len-1)*blockw), nextboxy+ezcenter(blockw,nextboxh), s, ONEFOURTHRAD, 0, 0, fTime);
}
static int getnextboxw(struct gamesettings* se){
	return (se->maxblocklen+1)*blockw;
}
static void drawboard(int origx, int origy, struct gboard* board, struct myblockshape* cursorshape, struct gamesettings* se, u64 fTime){
	int boardx=origx;
	int boardy=origy;
	offsettoactualboardcontentpos(&boardx,&boardy);
	int blocky=boardy+getboardblockoff(board, fTime);
	int bsize=__drawboard_bordersize();
	int deathminy=-1;
	if (fTime>=board->queuedUpshiftTime && board->queuedUpshiftTime!=0){
		u64 diff=tomltime(fTime-board->queuedUpshiftTime);
		deathminy=origy+(board->h*blockw+bsize*2)*(1-diff/(double)se->timetodie);
		if (deathminy<0){
			deathminy=0;
		}
	}
	// next box background
	int nextboxy=origy;
	int nextboxx=boardx+(board->w)*blockw;
	int nextboxw=(se->maxblocklen+1)*blockw;
	int nextboxh=blockw*2;
	for (int i=0;i<board->numnextpieces;++i){
		boarddrawrect(nextboxx,nextboxy+i*nextboxh,nextboxw,nextboxh,0,0,0,255,deathminy);
	}
	// cursor box background
	int cursorboxx=nextboxx;
	int cursorboxy=boardy+board->h*blockw+bsize-nextboxh;
	boarddrawrect(cursorboxx,cursorboxy,nextboxw,nextboxh,0,0,0,255,deathminy);
	//
	#ifndef BOARDIMAGEMAKER
	drawboardborder(origx, origy, board->w, board->h, deathminy);
	_boarddrawrect(boardx,boardy,board->w*blockw,board->h*blockw,92,92,255,255,100,0,0,255,deathminy);
	#endif
	if (board->upshiftEndTime){
		enableClipping(0, 0, getScreenWidth(), boardy+board->h*blockw);
	}
	int cursorboxpiecey=cursorboxy+ezcenter(blockw,nextboxh);
	ITERATENLIST(board->blocklist,{
			struct myplacedblock* b = _curnList->data;
			float bx;
			float by;
			double br;
			_getblockdrawprops(boardx, blocky, nextboxx, cursorboxpiecey, b, &bx, &by, &br, se, fTime);
			drawblockpart(bx, by, &b->s, br, b->startPopTime, se->poptime, fTime);
		});
	ITERATENLIST(board->blocklist,{
			struct myplacedblock* b = _curnList->data;
			float bx;
			float by;
			double br;
			_getblockdrawprops(boardx, blocky, nextboxx, cursorboxpiecey, b, &bx, &by, &br, se, fTime);
			drawheartpart(bx, by, &b->s, br, b->startPopTime, se->poptime, board->upshiftEndTime, board->deltaUpshift, board->h-b->y, b->isvertical, fTime);
		});
	if (board->upshiftEndTime){
		disableClipping();
	}
	// cursor block
	if (board->nextboxanimend){
		if (board->nextboxanimpiecesconsumed==1){
			enableClipping(cursorboxx, cursorboxy, nextboxw, nextboxh);
			cursorboxy+=partMoveEmpties(fTime, board->nextboxanimend, board->nextboxanimdelta, nextboxh);
			drawnextpieceat(cursorboxx, cursorboxy, cursorshape, nextboxw, nextboxh, fTime);
			disableClipping();
		}
	}else{
		drawnextpieceat(cursorboxx, cursorboxy, cursorshape, nextboxw, nextboxh, fTime);
	}
	// next box blocks
	enableClipping(nextboxx, nextboxy, nextboxw, nextboxh*board->numnextpieces);
	int i=0;
	if (board->nextboxanimend){
		nextboxy-=partMoveFills(fTime, board->nextboxanimend, board->nextboxanimdelta, nextboxh);
		nextboxy+=(board->nextboxanimpiecesconsumed-1)*nextboxh;
		i=-1;
	}
	for (;i<board->numnextpieces;++i){
		struct myblockshape* s;
		if (i==-1){
			s=cursorshape;
		}else{
			s=board->nextpieces[i];
		}
		drawnextpieceat(nextboxx, nextboxy, s, nextboxw, nextboxh, fTime);
		nextboxy+=nextboxh;
	}
	disableClipping();
	// draw chain popup
	int lastx=-1;
	ITERATENLIST(board->popups,{
			struct chainpopup* p = _curnList->data;
			char strbuff[6];
			if (p->ischain){
				strbuff[0]='x';
				sprintf(&strbuff[1],"%d",p->value);
			}else{
				sprintf(strbuff,"%d",p->value);
			}
			int x=p->x*blockw+boardx;
			int y=p->y*blockw+boardy-partMoveFills(fTime, p->timeend, se->poptime*POPUPTIMERATIO, blockw);
			int reqw=textWidth(mainfont,strbuff);
			int pad=blockw/4;
			int draww=intmax(reqw,mainFontHeight)+pad*2;
			if (lastx==p->x){
				y-=draww+pad;
			}else{
				lastx=p->x;
			}
			int textx=x+ezcenter(reqw,draww);
			int texty=y+ezcenter(mainFontHeight,draww);
			if (p->ischain){
				drawRectangle(x,y,draww,draww,255,255,255,255);
				drawborderrect(x, y, draww, draww, pad/2, 17,0,95,42,0,234, 255);
				gbDrawTextAlpha(mainfont, textx, texty, strbuff, 255,255,255,255);
			}else{
				gbDrawTextAlpha(mainfont, textx, texty, strbuff, 255,255,255,255);
			}
		});
}
static char __upshiftgenhelp_starcolorinlastrow(struct gboard* board, int x, enum blockcolor c){
	// returns if there's a star in the row above the one we're generating
	struct boardentry* rawent = board->board[x][board->h-2];
	if (rawent && (rawent->type & BOARDENT_STAR)){
		return ((struct boardentry_block*)rawent)->head->s.c==c;
	}
	return 0;
}
static char __upshiftgenhelp_isalreadycovered(int xposofinterest, struct nList* newblox){
	ITERATENLIST(newblox,{
			struct myplacedblock* b = _curnList->data;
			if (b->isvertical){
				if(b->x==xposofinterest){
					return 1;
				}
			}else{
				if (xposofinterest>=b->x && xposofinterest<b->x+b->s.len){
					return 1;
				}
			}
		});
	return 0;
}
static char __upshiftgenhelp_starofcolorisatx(int xposofinterest, int additionalyoff, struct nList* newblox, enum blockcolor c){
	ITERATENLIST(newblox,{
			struct myplacedblock* b = _curnList->data;
			if (b->s.c==c){
				if (b->isvertical){
					if (xposofinterest==b->x){
						int reqstaroff=additionalyoff+(b->s.len-1); // the "len" offset is for partial blocks. ones that just appeared should have length of 1.
						if (hasstaratoff(&b->s, reqstaroff)){
							return 1;
						}
					}
				}else if (additionalyoff==0){ // for horizontal blocks
					int reqstaroff=xposofinterest-b->x;
					// if in range
					if (!(reqstaroff<0 || reqstaroff>=b->s.len)){
						if (hasstaratoff(&b->s, reqstaroff)){
							return 1;
						}
					}
				}
			}
		});
	return 0;
}
// all the block has to do have at least one part of it be under one of these x positions
// returns 1 if it worked
static char __upshiftgenhelp_genblockthatcoversin(int minxcover, int maxxcover, struct nList** newblox, struct gboard* board, struct gamesettings* se, u64 fTime){
	// make a new block
	struct myplacedblock* newblock = malloc(sizeof(struct myplacedblock));
	initplacedblock(newblock);
	genblock(&newblock->s, se, fTime);
	newblock->isvertical = randbetween(0,5)==4;
	char didfail;
	int k;
	for (k=0;k<2;++k){ // retry with different orientation loop
		didfail=1;
		// init x position generator
		struct randrro* xrstate;
		int randxmin;
		int randxmax;
		if (newblock->isvertical){
			randxmin=minxcover;
			randxmax=maxxcover;
		}else{
			int leeway=(newblock->s.len-1);
			randxmin=limitNum(minxcover-leeway,0,INT32_MAX);
			randxmax=limitNum(maxxcover,0,board->w-leeway-1);
		}
		xrstate = newrandrangeonce(randxmin, randxmax);
		// try x positions
		while(didfail && (newblock->x=randrronext(xrstate))>=randxmin){
			// ensure that the desired space is free
			if (newblock->isvertical){
				if (__upshiftgenhelp_isalreadycovered(newblock->x, *newblox)){
					continue;
				}
			}else{
				int i;
				for (i=0;i<newblock->s.len;++i){
					if (__upshiftgenhelp_isalreadycovered(newblock->x+i, *newblox)){
						break;
					}
				}
				if (i!=newblock->s.len){
					continue;
				}
			}
			// okay, there is space for the block. time to try colors.
			// init color generator
			struct randrro* crstate = newrandrangeonce(COLOR_REAL, COLOR_REAL+se->numcolors-1);
			// try colors so that stars dont collide
			while(didfail && ((newblock->s.c=randrronext(crstate))>=COLOR_REAL)){
				didfail=0;
				if (newblock->isvertical){
					// first, check collision with blocks in the above row
					if (newblock->s.numhearts && newblock->s.heartloc[0]==0 && __upshiftgenhelp_starcolorinlastrow(board, newblock->x, newblock->s.c)){
						didfail=1;
					}else{
						for (int i=0;i<newblock->s.numhearts;++i){
							if (__upshiftgenhelp_starofcolorisatx(newblock->x-1, newblock->s.heartloc[i], *newblox, newblock->s.c) ||
								__upshiftgenhelp_starofcolorisatx(newblock->x+1, newblock->s.heartloc[i], *newblox, newblock->s.c)){
								didfail=1;
								break;
							}
						}
					}
				}else{
					// star collision with the row above
					for (int i=0;i<newblock->s.numhearts;++i){
						if (__upshiftgenhelp_starcolorinlastrow(board, newblock->x+newblock->s.heartloc[i], newblock->s.c)){
							didfail=1;
							break;
						}
					}
					if (!didfail){
						// for horizontal, star collision with other generated blocks is only on the ends.
						didfail=((hasstaratoff(&newblock->s, 0) &&  __upshiftgenhelp_starofcolorisatx(newblock->x-1, 0, *newblox, newblock->s.c)) ||
								 (hasstaratoff(&newblock->s, newblock->s.len-1) &&  __upshiftgenhelp_starofcolorisatx(newblock->x+newblock->s.len, 0, *newblox, newblock->s.c)));
					}
				}
			}
			freerandrro(crstate);
		}
		freerandrro(xrstate);
		if (!didfail){
			break;
		}
		newblock->isvertical=!newblock->isvertical;
	}
	if (k==2){
		#ifdef DEBUGOUT
		fprintf(stderr,"failed to place a bottom generated block in range %d;%d. very sad\n",minxcover,maxxcover);
		#endif
		freeplacedblock(newblock);
		free(newblock);
		return 0;
	}else{
		// happy add
		if (newblock->isvertical){
			newblock->s.partialleft=newblock->s.len-1;
			newblock->s.len=1;
		}
		newblock->y=board->h-1;
		addnList(newblox)->data=newblock;
		return 1;
	}
}
static struct nList* gennewbottomrow(struct gboard* board, struct gamesettings* se, u64 fTime){
	// generate the new bottom row
	char numblocksmade=0;
	int numoldbottomblocks=0;
	struct nList* newblox = NULL;
	// veritcal blocks need to be extended down, so just move them from the old list of bottom blocks to the new list.
	ITERATENLIST(board->lastbottomgen,{
			struct myplacedblock* b = _curnList->data;
			if (b->s.partialleft){
				b->s.partialleft--;
				b->s.len++;
				addnList(&(newblox))->data=b;
				ITERATENLISTREMOVETHISONE(board->lastbottomgen,0);
				numblocksmade++;
				numoldbottomblocks++;
			}
		});
	// now take all the other blocks and put a block under them to ensure they dont fall.
	// we dont care about the y position, that's garbage data. all we care about is x position and if it's veritcal
	ITERATENLIST(board->lastbottomgen,{
			struct myplacedblock* b = _curnList->data;
			numoldbottomblocks++;
			// decide what needs to be covered, and check if it already is
			int minxcover;
			int maxxcover; 
			if (b->isvertical){
				minxcover=b->x;
				maxxcover=minxcover;
			}else{
				minxcover=b->x;
				maxxcover=b->x+b->s.len-1;
			}
			numblocksmade+=__upshiftgenhelp_genblockthatcoversin(minxcover, maxxcover, &newblox, board, se, fTime);
		});
	if (numblocksmade<numoldbottomblocks){
		// find blocks that don't have anything under them
		ITERATENLIST(board->lastbottomgen,{
				struct myplacedblock* b = _curnList->data;
				char willfall;
				if (b->isvertical){ // and this is automatically not a partial block, by the way.
					willfall=!__upshiftgenhelp_isalreadycovered(b->x, newblox);
				}else{
					int i;
					for (i=0;i<b->s.len;++i){
						if (__upshiftgenhelp_isalreadycovered(b->x+i, newblox)){
							break;
						}
					}
					willfall=(i==b->s.len);
				}
				if (willfall){
					//#ifdef DEBUGOUT
					//fprintf(stderr,"block from x=%d will fall after upshift\n",b->x);
					//#endif
					struct myplacedblock* dupblock=dupplacedblock(b);
					dupblock->popcomboid=-10; // signal to not place the block
					addnList(&newblox)->data=dupblock;
				}
			});
	}
	if (numblocksmade<se->minimumupshiftblocksgenerated){
		int diff=se->minimumupshiftblocksgenerated-numblocksmade;
		for (int i=0;i<diff;++i){
			__upshiftgenhelp_genblockthatcoversin(0, board->w-1, &newblox, board, se, fTime);
		}
	}
	return newblox;
}
static void beginboardupshift(struct gboard* board, struct playercursor* cursor, struct gamesettings* se, u64 fTime){
	cursor->y--;
	if (cursor->y<cursor->das.miny){
		startcursormoveto(cursor, cursor->x, cursor->das.miny, fTime);
	}
	for (int i=0;i<board->w;++i){
		// shift a column column
		// because the top entry is always NULL, no need to save it.
		memmove(&(board->board[i][0]),&(board->board[i][1]),sizeof(struct boardentry*)*(board->h-1));
		board->board[i][board->h-1]=NULL;
	}
	ITERATENLIST(board->blocklist,{
			struct myplacedblock* b = _curnList->data;
			b->y--;
			// if it's a partial block, expand it.
			if (b->s.partialleft){
				struct boardentry_block* beb = malloc(sizeof(struct boardentry_block));
				beb->p.type=BOARDENT_BLOCKPART;
				beb->head=b;
				if (hasstaratoff(&b->s, b->s.len)){
					beb->p.type|=BOARDENT_STAR;
				}
				board->board[b->x][board->h-1]=(void*)beb;
				b->s.len++;
				b->s.partialleft--;
			}
		});
	board->deltaUpshift=se->upshiftanimtime;
	board->upshiftEndTime=fTime+tohdtime(board->deltaUpshift);
	////////////////////////////////////////////////
	struct nList* newblox;

	if (se->gengarbage){
		struct myplacedblock* b = malloc(sizeof(struct myplacedblock));
		initplacedblock(b);
		genblock(&b->s, se, fTime);
		free(b->s.heartloc);
		b->s.heartloc=NULL;
		b->s.numhearts=0;
		b->s.c=COLOR_REAL+maxcolors-1;
		b->isvertical=0;
		b->s.len=board->w;
		b->x=0;
		b->y=board->h-1;

		newblox=lowNewnList();
		newblox->data=b;
	}else{
		newblox=gennewbottomrow(board, se, fTime);
	}
	// free old bottom gen list
	ITERATENLIST(board->lastbottomgen,{
			freeplacedblock(_curnList->data);
		});
	freenList(board->lastbottomgen, 1);
	// place new ones them
	ITERATENLIST(newblox,{
			struct myplacedblock* b = _curnList->data;
			if ((!b->isvertical || b->s.len==1) && (b->popcomboid!=-10)){ // dont place the vertical extend ones and the ones that didnt get anything under them
				placeplacedblock(dupplacedblock(b), board, fTime);
			}
		});
	board->lastbottomgen=newblox;
}
///////////////////////////////////////////////
static int _updateifbigger(int* orig, int new){
	if (new>*orig){
		*orig=new;
		return 1;
	}
	return 0;
}
static void playablegamescoreresult(struct playablegame* g, struct scoreinfo* gotscore){
	if (!g->highscoredatapath){
		return;
	}
	char updateit=0;
	if (gotscore->score>g->highscores.score){
		g->highscores.score=gotscore->score;
		updateit=1;
	}
	{
		char doupdatetime;
		if (g->settings->isracemode){
			doupdatetime=(gotscore->timepassed<g->highscores.timepassed) || g->highscores.timepassed==0;
		}else{
			doupdatetime=(gotscore->timepassed>g->highscores.timepassed);
		}
		if (doupdatetime){
			g->highscores.timepassed=gotscore->timepassed;
			updateit=1;
		}
	}
	updateit|=_updateifbigger(&g->highscores.maxcombo, gotscore->maxcombo);
	updateit|=_updateifbigger(&g->highscores.maxchain, gotscore->maxchain);
	updateit|=_updateifbigger(&g->highscores.blockscleared, gotscore->blockscleared);
	if (updateit){
		FILE* fp = fopen(g->highscoredatapath,"wb");
		if (fp){
			if (writescorefile(fp, &g->highscores)){
				sadflash();
				fclose(fp);
				fp = fopen(g->highscoredatapath,"wb"); // try to zero the file
				if (fp){
					fclose(fp);
				}
			}else{
				fclose(fp);
			}
		}else{
			perror(g->highscoredatapath);
			sadflash();
		}
	}
}
static void initplayablegame(struct playablegame* g){
	g->freeplayable=NULL;
	g->highscoredatapath=NULL;
	g->music=NULL;
	g->musicstarted=0;
}
static void alwaysrunfreeplayable(struct playablegame* playable){
	if (playable->music){
		stopMusic(playable->musichandle);
		freeMusic(playable->music); // assume it's safe to free playing music
	}
}
// return 1 if we want to play again
// return 2 to quit
static int _playgamewithstate(struct playablegame* playable, struct gamestate* state){
	playable->initialfunc(state);
	while(1){
		state->fTime=getHDTime();
		int updateret=playable->updatefunc(state);
		if (updateret){
			return updateret;
		}
		startDrawing();
		playable->drawfunc(state);
		endDrawing();
	}
}
static int playgame(struct playablegame* playable){
	if (playable->highscoredatapath){
		FILE* fp = fopen(playable->highscoredatapath,"rb");
		if (fp){
			if (loadscorefile(fp, &playable->highscores)){
				initscoreinfo(&playable->highscores);
				sadflash();
			}
			fclose(fp);
		}else{
			initscoreinfo(&playable->highscores);
		}
	}
	int playgameret=1;
	while(playgameret==1){
		struct gamestate* state = playable->newstate(playable);
		state->game=playable;
		state->fTime=0;
		playgameret=_playgamewithstate(playable, state);
		playable->freestate(state);
	}
	alwaysrunfreeplayable(playable);
	if (playable->freeplayable){
		playable->freeplayable(playable);
	}
	return playgameret;
}
////////////////////////////////////////////////////////////
// MODE - GENERIC
////////////////////////////////////////////////////////////
struct gamestate_generic{
	struct gamestate p;
	struct gboard* board;
	struct playercursor* cursor;
	u64 pauseStartTime;
	u64 timeOff;
	u64 boardanimend;
	char boardanimdir; // 0 for intro, 1 for outro
	char retryafteroutro;
	char hidecursor;
	// a bitmap
	// 1 - play the win sound
	// 2 - DO NOT send the score.
	// 4 - quit to the title screen after. dont show menu.
	char deathaction;
};
#define GAMESTATE_GENERIC_OUTROANIMLEN 1000
static void _modegeneric_introanim(struct gamestate* _g, char direction){
	struct gamestate_generic* g = (void*)_g;
	g->boardanimdir=direction;
	g->boardanimend=getHDTime()+tohdtime(GAMESTATE_GENERIC_OUTROANIMLEN);
	while((_g->fTime=getHDTime())<g->boardanimend){
		startDrawing();
		_g->fTime=getHDTime();
		_g->game->drawfunc(_g);
		endDrawing();
		controlsStart();
		if (menuselectpressed()){
			controlsEnd();
			break;
		}
		controlsEnd();
	}
	g->boardanimend=0;
}
#define GAMESTATE_GENERIC_COUNTDOWNTIME 850
#define GAMESTATE_GENERIC_NUMCOUNTDOWN 3
static void modegeneric_initial(struct gamestate* _g){
	struct gamestate_generic* g = (void*)_g;
	g->board->queuedUpshiftTime=0;
	g->hidecursor=1;
	_modegeneric_introanim((void*)_g, 0);
	int forcecursorx=g->cursor->x;
	int forcecursory=g->cursor->y;
	// do countdown
	{
		u64 countdownstart=getHDTime();
		//GAMESTATE_GENERIC_COUNTDOWNTIME
		int beepsplayed=0;
		//while((_g->fTime=getHDTime())<countdownend){
		while(1){
			u64 fTime=getHDTime();
			int curbeep=tomltime(fTime-countdownstart)/GAMESTATE_GENERIC_COUNTDOWNTIME;
			if (curbeep>=GAMESTATE_GENERIC_NUMCOUNTDOWN){
				break;
			}
			controlsStart();
			if (processcursordas(g->cursor, &g->cursor->das, g->p.game->controls, fTime)){
				g->cursor->x=forcecursorx;
				g->cursor->y=forcecursory;
			}
			if (wasJustPressed(BUTTON_START)){
				controlsEnd();
				break;
			}
			controlsEnd();
			if (beepsplayed<=curbeep){
				playSound(countdownSE[0],0);
				beepsplayed++;
			}
			startDrawing();
			_g->fTime=fTime;
			_g->game->drawfunc(_g);
			int bx,by,bw,bh;
			stdboarddraw(g->board, g->p.game->settings, &bx, &by, &bw, &bh);
			gbDrawTextfCenter(mainfont, bx, by+ezcenter(mainFontHeight,bh), 0,0,0,255, getboardwnonextbox(g->board), "%d", 4-beepsplayed);
			endDrawing();
		}
		playSound(countdownSE[1],0);
	}
	g->hidecursor=0;
	if (!g->p.game->musicstarted && g->p.game->music){
		g->p.game->musicstarted=1;
		playMusic(g->p.game->music,0);
	}
	// finally, set the variables for a proper game start
	boardongamestart(g->board, g->p.game->settings, getHDTime());
}
#define GAMEOVERMENUDELAY 700 // sync this to "Game over" sound time?
// 1 - retry
// 2 - quit
static int _modegeneric_gameovermenu(struct gamestate_generic* g){
	u64 whenmenuappear=getHDTime()+tohdtime(GAMEOVERMENUDELAY);
	struct pumenu bla;
	char* wow[]={"Retry","Quit"};
	char menuon=0;
	while(1){
		u64 fTime=getHDTime();
		controlsStart();
		if (menuon){
			if (pumenucontrols(&bla,fTime)!=-1){
				controlsEnd();
				break;
			}
		}else if (fTime>=whenmenuappear){
			initpumenu(&bla);
			menuon=1;
			bla.options=wow;
			bla.numoptions=sizeof(wow)/sizeof(char*);
			bla.choice=0;
			bla.introtime=1000;
		}else if (menuselectpressed()){
			whenmenuappear=fTime;
		}
		controlsEnd();
		startDrawing();
		g->p.game->drawfunc((void*)g);
		if (menuon){
			drawpumenu(&bla,fTime);
		}
		endDrawing();
	}
	return bla.choice+1;
}
static void modegeneric_draw(struct gamestate* _g){
	struct gamestate_generic* g = (void*)_g;
	int bw,bh,bx,by;
	stdboarddraw(g->board, g->p.game->settings, &bx, &by, &bw, &bh);
	u64 fTime;
	if (g->board->diedat){
		fTime=g->board->diedat;
	}else{
		fTime=g->p.fTime;
	}
	drawbackground();
	if (g->boardanimend){
		int screenheight=getScreenHeight();
		int offset=partMoveFills(g->p.fTime, g->boardanimend, GAMESTATE_GENERIC_OUTROANIMLEN, screenheight);
		if (g->boardanimdir==1){
			by-=offset;
		}else{
			by+=(screenheight-offset);
		}
		drawboard(bx,by,g->board,g->cursor->block,g->p.game->settings,fTime);
	}else{
		if (!g->pauseStartTime){
			drawboard(bx,by,g->board,g->cursor->block,g->p.game->settings,fTime);
		}else{
			_drawlabelat(bx+getboardwnonextbox(g->board)/2, by+ezcenter(mainFontHeight,bh), blockw/3, 255, 255, "Paused");
		}
	}
	if (!g->hidecursor){
		drawcursor(bx, by, g->cursor, g->board, fTime);
	}
}
static int modegeneric_update(struct gamestate* _g){
	struct gamestate_generic* g = (void*)_g;
	resetgsound();
	u64 origftime=g->p.fTime;
	if (g->pauseStartTime){
		g->p.fTime=g->pauseStartTime;
	}
	g->p.fTime-=g->timeOff;
	controlsStart();
	if (wasJustPressed(BUTTON_START)){
		if (g->pauseStartTime){
			g->timeOff+=(origftime-g->pauseStartTime);
			g->pauseStartTime=0;
		}else if (!g->board->diedat){
			if (_g->game->settings->isracemode){
				playSound(diesound,0);
				g->board->diedat=g->p.fTime;
			}else{
				g->pauseStartTime=origftime;
			}
		}
		g->hidecursor=(g->pauseStartTime!=0);
	}
	if (!g->pauseStartTime){
		playerupdateboard(g->board, g->cursor, g->p.game->controls, g->p.game->settings, g->p.fTime);
		if (g->board->diedat){
			controlsEnd();
			g->hidecursor=1;
			if ((g->deathaction & 2)==0){
				playablegamescoreresult(g->p.game, &g->board->score);
			}
			if ((g->deathaction & 1)==0){
				playSound(diesound,0);
			}else{
				playSound(winsound,0);
			}
			int ret;
			if (g->deathaction & 4){
				ret=3;
			}else{
				ret=_modegeneric_gameovermenu(g);
			}
			g->retryafteroutro=(ret==1);
			if (!g->retryafteroutro){
				fadeoutMusic(g->p.game->musichandle,GAMESTATE_GENERIC_OUTROANIMLEN*.9);
			}
			_modegeneric_introanim((void*)_g, 1);
			return ret;
		}
	}
	playgsound();
	controlsEnd();
	return 0;
}
static void _modegeneric_newstate(struct playablegame* g, struct gamestate_generic* state){
	u64 fTime=getHDTime();
	state->hidecursor=0;
	state->boardanimend=0;
	state->boardanimdir=0;
	state->pauseStartTime=0;
	state->timeOff=0;
	state->deathaction=0;
	state->retryafteroutro=0;
	state->board=malloc(sizeof(struct gboard));
	newboard(state->board,g->settings->defaultboardw,g->settings->defaultboardh,g->settings);
	generateinitialboard(state->board,g->settings,fTime);
	
	state->cursor=malloc(sizeof(struct playercursor));
	newplayercursor(state->cursor, g->settings, state->board, boardgivenextpiece(state->board, g->settings, fTime), fTime);
}
static struct gamestate* modegeneric_newstate(struct playablegame* g){
	struct gamestate_generic* state = malloc(sizeof(struct gamestate_generic));
	_modegeneric_newstate(g, state);
	return (void*)state;
}
static void _modegeneric_freestate(struct gamestate* _g){
	struct gamestate_generic* g = (void*)_g;
	freeboard(g->board);
	free(g->board);
	freeplayercursor(g->cursor);
	free(g->cursor);
}
static void modegeneric_freestate(struct gamestate* _g){
	_modegeneric_freestate(_g);
	free(_g);
}
static void playable_setgeneric(struct playablegame* game, struct gamesettings* settings, struct pcontrols* controls, crossMusic* music){
	initplayablegame(game);
	game->settings=settings;
	game->controls=controls;
	game->initialfunc=modegeneric_initial;
	game->drawfunc=modegeneric_draw;
	game->updatefunc=modegeneric_update;
	game->newstate=modegeneric_newstate;
	game->freestate=modegeneric_freestate;
	game->music=music;
	if (!game->music){
		game->music=loadvpkmusic("assets/se/puzzlemaster.ogg");
		setMusicVolumeBefore(game->music,128*.5);
	}
}
////////////////////////////////////////////////////////////
// MODE - GENERIC WITH LABELS
////////////////////////////////////////////////////////////
/////// Internal data types
struct playableextra_labels{
	int numlabels;
	char** labeltext;
	crossTexture* logo;
	void(*updatelabels)(struct gamestate* g);
	void* extradata; // will be freed. but doesnt have its own free function
	char introisdone;
	char disabled;
};

struct playablegame_labels{
	struct playablegame p;
	struct playableextra_labels labels;
};

static void initplayableextra_labels(struct playableextra_labels* dest){
	dest->numlabels=0;
	dest->labeltext=NULL;
	dest->logo=NULL;
	dest->extradata=NULL;
	dest->introisdone=0;
	dest->disabled=0;
}
static void _freejustlabeltext(struct playableextra_labels* l){
	for (int i=0;i<l->numlabels;++i){
		free(l->labeltext[i]);
	}
	free(l->labeltext);
	l->numlabels=0;
}
/// functions
static int modelabels_update(struct gamestate* _g){
	int ret=modegeneric_update(_g);
	struct playableextra_labels* lab = &((struct playablegame_labels*)_g->game)->labels;
	lab->updatelabels(_g);
	return ret;
}
static void _getlabelpos(struct playableextra_labels* lab, struct gboard* board, struct gamesettings* se, int* retstampsize, int* smidx, int* sfirsty, int* stampyoff, int* logox, int* logoy, int* logow, int* logoh){
	int bw,bh,bx,by;
	stdboarddraw(board, se, &bx, &by, &bw, &bh);
	
	int stampminx=bx+bw;
	int stampareaw=getScreenWidth()-stampminx;
	int stampareah=getScreenHeight()*.8;
	if (smidx){
		*smidx=stampminx+(stampareaw)/2;

		int stampsize=blockw/3;
		char canretrystampcalc=1;
	retrystampcalc:;
		struct stampprops props;
		calcstamp(&props, blockw, mainFontHeight, stampsize);
		*sfirsty=ezcenter(props.entireh*lab->numlabels+stampsize*(lab->numlabels-1),stampareah);
		if (*sfirsty<0 && canretrystampcalc){
			stampsize=intmax(1,blockw/4);
			canretrystampcalc=0;
			goto retrystampcalc;
		}
		*stampyoff=props.entireh+stampsize;
		*retstampsize=stampsize;
	}
	if (lab->logo){
		*logoh=blockw*2.333;
		// if fitting to width gives a smaller height
		int altw=stampareaw-blockw/10;
		int alth=getscaledh(altw, lab->logo);
		if (alth<*logoh){
			*logow=altw;
			*logoh=alth;
		}else{
			*logow=getscaledw(*logoh, lab->logo);
		}
		*logox=stampminx+ezcenter(*logow,stampareaw);
		*logoy=stampareah+ezcenter(*logoh,getScreenHeight()-stampareah);
	}else{
		*logox=stampminx;
		*logoy=stampareah;
	}
}
static void _drawlabelat(int middlex, int y, int stampsize, unsigned char stampa, unsigned char texta, const char* text){
	int reqw=textWidth(mainfont, text);
	struct stampprops props;
	calcstamp(&props, reqw, mainFontHeight, stampsize);
	int drawx=middlex-props.entirew/2;
	drawstamp(drawx, y, &props, stampa, stampres);
	if (texta){
		gbDrawTextAlpha(mainfont, drawx+props.cornerw+ezcenter(reqw,props.insidew), y+props.cornerw+ezcenter(mainFontHeight,props.insideh), text, 0,0,0, texta);
	}
}
static void _modelabels_drawlabels(struct gamestate_generic* g){
	struct playableextra_labels* lab = &((struct playablegame_labels*)g->p.game)->labels;
	if (lab->disabled){
		return;
	}
	char isoutro=g->boardanimend && g->boardanimdir && !g->retryafteroutro;
	unsigned char a=255;
	if (isoutro){
		a=partMoveEmpties(g->p.fTime, g->boardanimend, GAMESTATE_GENERIC_OUTROANIMLEN, 255);
	}
	int sx,sy,syoff,stampsize;
	int lx,ly,lw,lh;
	_getlabelpos(lab, g->board, g->p.game->settings, &stampsize, &sx, &sy, &syoff, &lx, &ly, &lw, &lh);
	for (int i=0;i<lab->numlabels;++i){
		_drawlabelat(sx, sy, stampsize, a, a, lab->labeltext[i]);
		sy+=syoff;
	}
	// draw logo
	if (lab->logo){
		if (isoutro){
			ly+=partMoveFills(g->p.fTime, g->boardanimend, GAMESTATE_GENERIC_OUTROANIMLEN,getScreenHeight()-ly);
		}
		drawTextureSized(lab->logo,lx,ly,lw,lh);
	}
}
static void _modelabels_labelintro(struct gamestate_generic* g, struct playableextra_labels* lab){
	lab->introisdone=1;

	// ensure labels fit
	{
		int sx,sy,syoff,stampsize;
		int lx,ly,lw,lh;
		_getlabelpos(lab, g->board, g->p.game->settings, &stampsize, &sx, &sy, &syoff, &lx, &ly, &lw, &lh);
		for (int i=0;i<lab->numlabels;++i){
			int reqw=textWidth(mainfont, lab->labeltext[i]);
			struct stampprops props;
			calcstamp(&props, reqw, mainFontHeight, stampsize);
			if (sx+props.entirew/2>=getScreenWidth()){
				lab->disabled=1;
				return;
			}
		}
	}
	
	int putseplayed=0;
	crossSE* putse = loadputstampse();
	mltime putstamptime=400;
	mltime stamptexttime=100; // the first (this) number of milli during the next stamp, the text appears.
	mltime totaltime=putstamptime*lab->numlabels+stamptexttime;
	u64 animstarttime=getHDTime();
	while(1){
		u64 fTime=getHDTime();
		mltime curdiff=tomltime(fTime-animstarttime);
		int stampsdone=curdiff/putstamptime;
		mltime thisstamp=curdiff%putstamptime;
		if (stampsdone>=lab->numlabels){
			if (thisstamp>stamptexttime){
				break;
			}
		}
		if (putseplayed<stampsdone){
			playSound(putse, 0);
			putseplayed++;
		}
		char doprevtext=(thisstamp<stamptexttime);
		
		int sx,sy,syoff,stampsize;
		int lx,ly,lw,lh;
		_getlabelpos(lab, g->board, g->p.game->settings, &stampsize, &sx, &sy, &syoff, &lx, &ly, &lw, &lh);
		
		startDrawing();
		drawbackground();
		if (lab->logo){
			drawTextureSizedAlpha(lab->logo, lx, ly, lw, lh, 255*curdiff/(double)totaltime);
		}
		int i;
		for (i=0;i<stampsdone;++i){
			int a=255;
			if (i==stampsdone-1 && doprevtext){
				a=(curdiff/(double)stamptexttime)*255;
			}
			_drawlabelat(sx, sy, stampsize, 255, a, lab->labeltext[i]);
			sy+=syoff;
		}
		if (stampsdone<lab->numlabels){
			if (thisstamp<putstamptime){
				double scale=5-4*(thisstamp/(double)putstamptime);
				int reqw=textWidth(mainfont, lab->labeltext[i]);
				struct stampprops props;
				calcstamp(&props, reqw, mainFontHeight, stampsize);
				// because the cornerw is an integer, the animation is a bit choppy. we make it super fast so nobody will notice
				props.cornerw*=scale;
				int circlew=props.cornerw*(13/(double)20);
				int rectw=props.cornerw*(7/(double)20);
				int edgepartw=circlew+rectw;
				props.insidew=edgepartw*props.__numedgex+rectw;
				props.insideh=edgepartw*props.__numedgey+rectw;
				props.entirew=props.insidew+props.cornerw*2;
				props.entireh=props.insideh+props.cornerw*2;
				int drawx=sx-props.entirew/2;
				drawstamp(drawx, sy, &props, 255, stampres);
			}
		}
		endDrawing();
		controlsStart();
		if (menuselectpressed()){
			controlsEnd();
			break;
		}
		controlsEnd();
	}
	freeSound(putse);
}
static void modelabels_initial(struct gamestate* _g){
	struct playableextra_labels* lab = &((struct playablegame_labels*)_g->game)->labels;
	// initialize labels. even if we already did the label intro, they need updates for the new board
	lab->updatelabels((void*)_g);
	if (!lab->introisdone){
		_modelabels_labelintro((void*)_g, lab);
	}
	modegeneric_initial(_g);
}
static void modelabels_draw(struct gamestate* _g){
	modegeneric_draw(_g);
	_modelabels_drawlabels((void*)_g);
}
static void modelabels_freeplayable(struct playablegame* p){
	struct playableextra_labels* lab = &((struct playablegame_labels*)p)->labels;
	_freejustlabeltext(lab);
	if (lab->logo){
		freeTexture(lab->logo);
	}
	free(lab->extradata);
}
// labelsmap[0] is number of lables. rest is from (enum modelabelsmap) in order.
static void playable_setlabelsmode(struct playablegame_labels* game, struct gamesettings* settings, struct pcontrols* controls){
	playable_setgeneric((struct playablegame*)game, settings, controls, NULL);
	game->p.drawfunc=modelabels_draw;
	game->p.updatefunc=modelabels_update;
	game->p.initialfunc=modelabels_initial;
	
	// set up label handler
	initplayableextra_labels(&game->labels);
	game->p.freeplayable=modelabels_freeplayable;
}
////////////////////////////////////////////////////////////
// MODE - EASY LABELS
////////////////////////////////////////////////////////////
enum modelabelsmap{
	MODELABEL_SCORE=1,
	MODELABEL_HIGHSCORE,
	MODELABEL_TIME,
	MODELABEL_HIGHTIME,
	MODELABEL_BLOCKSCLEARED,
	MODELABEL_CURSPEED,
	MODELABEL_MAXCHAIN,
	MODELABEL_BESTCHAIN,
	MODELABEL_MAXCOMBO,
	MODELABEL_BESTCOMBO,
	MODELABEL_NULL, // you must fill it in later.
};
static char* maketimestring(mltime timepassed, const char* prefixstr){
	int msec=(timepassed%1000)/100;
	int sec=timepassed/1000;
	int min=sec/60;
	sec=sec%60;
	return easySprintf("%s: %02d:%02d.%d",prefixstr,min,sec,msec);
}
static void modeeasylabels_updatelabels(struct gamestate* _g){
	struct playableextra_labels* lab = &((struct playablegame_labels*)_g->game)->labels;
	struct gamestate_generic* g=(void*)_g;
	_freejustlabeltext(lab);
	char* map=lab->extradata;
	lab->numlabels=map[0];
	lab->labeltext=malloc(sizeof(char*)*(lab->numlabels));
	for (int i=1;i<=map[0];++i){
		struct scoreinfo* info=&(g->board->score);
		char* desttext;
		switch(map[i]){
			case MODELABEL_SCORE:
				desttext=easySprintf("Score: %"SCOREDIGITS""PRIu64,info->score);
				break;
			case MODELABEL_HIGHSCORE:
				desttext=easySprintf("High: %"SCOREDIGITS""PRIu64,g->p.game->highscores.score);
				break;
			case MODELABEL_HIGHTIME:;
				desttext=maketimestring(g->p.game->highscores.timepassed, _g->game->settings->isracemode ? "Fastest" : "Longest");
				break;
			case MODELABEL_TIME:;
				desttext=maketimestring(info->timepassed, "Time");
				break;
			case MODELABEL_BLOCKSCLEARED:
				desttext=easySprintf("Block: %d",info->blockscleared);
				break;
			case MODELABEL_CURSPEED:
				desttext=easySprintf("Speed: %.2f",getspeedlv(getnextupshifttime(g->p.game->settings,g->board->score.timepassed)));
				break;
			case MODELABEL_MAXCHAIN:
				desttext=easySprintf("Max Chain: %d",info->maxchain);
				break;
			case MODELABEL_BESTCHAIN:
				desttext=easySprintf("Best Chain: %d",g->p.game->highscores.maxchain);
				break;
			case MODELABEL_MAXCOMBO:
				desttext=easySprintf("Max Combo: %d",info->maxcombo);
				break;
			case MODELABEL_BESTCOMBO:
				desttext=easySprintf("Best Combo: %d",g->p.game->highscores.maxcombo);
				break;
			default:
			case MODELABEL_NULL:
				desttext="NULL";
				break;
		}
		lab->labeltext[i-1]=desttext;
	}
}
// labelsmap[0] is number of lables. rest is from (enum modelabelsmap) in order.
static void playable_seteasylabelsmode(struct playablegame_labels* game, struct gamesettings* settings, struct pcontrols* controls, const char* labelsmap, const char* logopath){
	playable_setlabelsmode(game, settings, controls);
	if (!labelsmap){
		return;
	}
	// set up label handler
	struct playableextra_labels* labelstruct = &game->labels;
	labelstruct->updatelabels=modeeasylabels_updatelabels;
	labelstruct->extradata=malloc(labelsmap[0]+1);
	memcpy(labelstruct->extradata,labelsmap,labelsmap[0]+1);
	if (logopath){
		labelstruct->logo=loadvpkpng(logopath);
	}
}
////////////////////////////////////////////////////////////
// MODE - DEATH
////////////////////////////////////////////////////////////
static void modedeath_updatelabels(struct gamestate* _g){
	struct playableextra_labels* lab = &((struct playablegame_labels*)_g->game)->labels;
	struct gamestate_generic* g=(void*)_g;
	_freejustlabeltext(lab);
	lab->numlabels=2;
	lab->labeltext=malloc(sizeof(char*)*(lab->numlabels));
	lab->labeltext[0]=maketimestring(g->board->score.timepassed, "Struggled");
	lab->labeltext[1]=maketimestring(g->p.game->highscores.timepassed, "Best");
}
static void playable_setdeathmode(struct playablegame_labels* game, struct gamesettings* settings, struct pcontrols* controls){
	playable_setlabelsmode(game, settings, controls);
	struct playableextra_labels* labelstruct = &game->labels;
	labelstruct->updatelabels=modedeath_updatelabels;
	labelstruct->logo=loadvpkpng("assets/img/ui/gamemodelogo/death.png");
}
////////////////////////////////////////////////////////////
// MODE - Race
////////////////////////////////////////////////////////////
static int moderace_update(struct gamestate* _g){
	int ret=modelabels_update((void*)_g);
	struct gamestate_generic* g = (void*)_g;
	if (!g->board->diedat){
		g->deathaction=2;
		if (g->board->blocklist==NULL){
			g->board->diedat=_g->fTime;
			g->deathaction=1;
		}
	}
	return ret;
}
static void playable_setracemode(struct playablegame_labels* game, struct gamesettings* settings, struct pcontrols* controls, const char* logopath){
	char labelmap[] = {
		4,
		MODELABEL_TIME,
		MODELABEL_HIGHTIME,
		MODELABEL_BLOCKSCLEARED,
		MODELABEL_MAXCHAIN,
	};
	playable_seteasylabelsmode(game, settings, controls, labelmap, logopath);
	game->p.updatefunc=moderace_update;
}
////////////////////////////////////////////////////////////
// MODE - TUTORIAL
////////////////////////////////////////////////////////////
struct playablegame_tutorial{
	struct playablegame p;
	char introdone;
	crossSE* putse;
};
struct gamestate_tutorial{
	struct gamestate_generic p;
	int progress;
	char** curlines;
	int numlines;
	//int extravar;
	const char* nexttext;
	const char* nextimagefname;
	u64 stampendtime;
	crossTexture* helpimage;
};
static void _modetutorial_calc(struct gboard* board, struct gamesettings* se, int* sx, int* sy, int* stampsize, struct stampprops* retprops, int* wrapw){
	int bw,bh,bx,by;
	stdboarddraw(board, se, &bx, &by, &bw, &bh);
	int pad=blockw/2;
	int lstampsize=blockw/3;
	struct stampprops s;
	calcstamp(&s, getScreenWidth()-bx-bw-pad*2-lstampsize, getScreenHeight()-pad*2-lstampsize, lstampsize);
	if (sx){
		int _xoff=bx+bw;
		int lsx=_xoff+ezcenter(s.entirew,getScreenWidth()-_xoff);
		int lsy=ezcenter(s.entireh,getScreenHeight());
		*sx=lsx;
		*sy=lsy;
	}
	if (stampsize){
		*stampsize=blockw/3;
		*retprops=s;
	}
	if (wrapw){
		*wrapw=s.insidew;
	}
}
#define TUTORIALPUTSTAMPTIME 130
static void modetutorial_drawstampanim(struct gamestate_generic* g, u64 animendtime, u64 fTime){
	double scale=1;
	if (fTime<animendtime){
		scale=2-partMoveFills(fTime, animendtime, TUTORIALPUTSTAMPTIME, 1);
	}
	struct stampprops props;
	int sx,sy,stampsize;
	_modetutorial_calc(g->board, g->p.game->settings, &sx, &sy, &stampsize, &props, NULL);
	int oldw=props.entirew;
	int oldh=props.entireh;
	props.cornerw*=scale;
	int circlew=props.cornerw*(13/(double)20);
	int rectw=props.cornerw*(7/(double)20);
	int edgepartw=circlew+rectw;
	props.insidew=edgepartw*props.__numedgex+rectw;
	props.insideh=edgepartw*props.__numedgey+rectw;
	props.entirew=props.insidew+props.cornerw*2;
	props.entireh=props.insideh+props.cornerw*2;
	drawstamp(sx-(props.entirew-oldw)/2, sy-(props.entireh-oldh)/2, &props, 255, stampres);
}
static void _modetutorial_queuemessage(struct gamestate_tutorial* t, u64 fTime, const char* newmsg){
	t->nexttext=newmsg;
	t->stampendtime=fTime+tohdtime(TUTORIALPUTSTAMPTIME);
}
static void _modetutorial_freelines(struct gamestate_tutorial* t){
	for (int i=0;i<t->numlines;++i){
		free(t->curlines[i]);
	}
	free(t->curlines);
	if (t->helpimage){
		freeTexture(t->helpimage);
		t->helpimage=NULL;
	}
}
static void _modetutorial_settext(struct gamestate_tutorial* g, const char* msg){
	_modetutorial_freelines(g);
	int wrapw;
	_modetutorial_calc(g->p.board, g->p.p.game->settings, NULL, NULL, NULL, NULL, &wrapw);
	wrapText(msg, &g->numlines, &g->curlines, wrapw, mainfont);
}
static void modetutorial_initial(struct gamestate* _g){
	struct playablegame_tutorial* g2=(void*)_g->game;
	struct gamestate_tutorial* t = (void*)_g;
	if (!g2->introdone){
		g2->introdone=1;
		struct gamestate_generic* g = (void*)_g;
		_modetutorial_queuemessage(t, getHDTime(), "");
		t->nexttext=NULL;
		while(1){
			u64 fTime=getHDTime();
			if (fTime>t->stampendtime){
				playSound(((struct playablegame_tutorial*)(g->p.game))->putse,0);
				break;
			}
			startDrawing();
			drawbackground();
			modetutorial_drawstampanim(g,t->stampendtime, fTime);
			endDrawing();
			controlsReset();
		}
	}
	_modetutorial_settext((void*)_g,
						  "Welcome to Happy Blocks.\n\n\
You can move the cursor woth D-pad.\n\n\
To continue, move your cursor to the top right of the board.");
	modegeneric_initial(_g);
}
static void modetutorial_freestate(struct gamestate* _g){
	struct gamestate_tutorial* t = (void*)_g;
	_modetutorial_freelines(t);
	modegeneric_freestate(_g);
}
static struct gamestate* modetutorial_newstate(struct playablegame* g){
	struct gamestate_tutorial* ret = malloc(sizeof(struct gamestate_tutorial));
	_modegeneric_newstate(g, (void*)ret);
	ret->progress=0;
	ret->curlines=NULL;
	ret->numlines=0;
	ret->nexttext=NULL;
	ret->stampendtime=0;
	ret->helpimage=NULL;
	return (void*)ret;
}
static void modetutorial_draw(struct gamestate* _g){
	modegeneric_draw(_g);
	struct gamestate_generic* gen=(void*)_g;
	char isoutro=gen->boardanimend && gen->boardanimdir && !gen->retryafteroutro;
	unsigned char a=255;
	if (isoutro){
		a=partMoveEmpties(gen->p.fTime, gen->boardanimend, GAMESTATE_GENERIC_OUTROANIMLEN, 255);
	}

	/* int rectpad=blockw/5; */
	/* drawborderrect(lx-rectpad,ly-rectpad,customnamew+rectpad*2,mainFontHeight+rectpad*2,blockw*.1,0,129,27,0,236,50,a); */
	/* int dropshadowoff=mainFontHeight/10; */
	/* gbDrawTextAlpha(mainfont, lx+dropshadowoff, ly+dropshadowoff, custommodename, 0,0,0,a); */
	/* gbDrawTextAlpha(mainfont, lx, ly, custommodename, 255,255,255,a); */
	struct gamestate_tutorial* g = (void*)_g;
	int sx,sy,stampsize;
	int linew;
	struct stampprops p;
	_modetutorial_calc(g->p.board, g->p.p.game->settings, &sx, &sy, &stampsize, &p, &linew);
	drawstamp(sx, sy, &p, a, stampres);
	for (int i=0;i<g->numlines;++i){
		gbDrawTextAlpha(mainfont, sx+stampsize, sy+stampsize+i*mainFontHeight, g->curlines[i], 0,0,0,a);
	}
	if (g->helpimage){
		int y=sy+stampsize+g->numlines*mainFontHeight+blockw/2;
		int maxh=getScreenHeight()-y-stampsize;
		int w,h;
		fitinbox(g->helpimage,linew-blockw,maxh,&w,&h);
		drawTextureSizedAlpha(g->helpimage,sx+blockw/2+stampsize+ezcenter(w,linew-blockw),y+ezcenter(h,maxh),w,h,a);
	}

	if (g->nexttext){
		modetutorial_drawstampanim(gen, g->stampendtime, gen->p.fTime);
	}
	
	//_modelabels_drawlabels((void*)_g);
}
static int modetutorial_update(struct gamestate* _g){
	int ret=modegeneric_update(_g);
	struct gamestate_tutorial* g = (void*)_g;
	u64 fTime=g->p.p.fTime;
	if (g->nexttext){
		if (fTime>g->stampendtime){
			crossTexture* newimg=NULL;
			if (g->nextimagefname){
				char* fullpath=easyCombineStrings(3,getFixPathString(TYPE_EMBEDDED),"assets/img/tut/",g->nextimagefname);
				newimg=loadPNG(fullpath);
				free(fullpath);
			}
			playSound(((struct playablegame_tutorial*)g->p.p.game)->putse,0);
			_modetutorial_settext((void*)g, g->nexttext);
			g->helpimage=newimg;
			g->nexttext=NULL;
			g->nextimagefname=NULL;
		}
	}else{
		const char* newtext=NULL;
		const char* newimg=NULL;
		if (g->progress==0){
			if (g->p.cursor->y==g->p.cursor->das.miny && g->p.cursor->x==g->p.cursor->das.maxx){
				newtext="In this game, you place blocks.\n\nThe color and shape of your next block is shown on the cursor and in the bottom right.\n\nTo rotate your block, press X or O.\n\nTo put your block on the board, press Square or Triangle.\n\nThe queue of blocks you will recieve next is shown in the top right.\n\nTo continue the tutorial, get 3 blocks on your board.";
				if (nListLen(g->p.board->blocklist)>=3){
					g->progress=1;
				}
				//newtext="Each block has one or more \"Cores\" on it represented by a spinning shape.\nTo clear blocks, align them so that two Cores of the same color touch.\n\nTo continue, clear 6 blocks.";
				//g->p.board->score.blockscleared=0;
			}
		}
		if (g->progress==1){
			if (nListLen(g->p.board->blocklist)>=3){
				g->p.board->score.blockscleared=0;
				newtext="Each block has one or more \"Cores\" on it represented by a spinning shape.\nTo clear blocks, align them so that two Cores of the same color touch.\n\nTo continue, clear 6 blocks.";
				newimg="match.png";
			}
		}else if (g->progress==2){
			if (g->p.board->score.blockscleared>=6){
				g->p.board->score.maxcombo=0;
				newtext="If you clear a set of blocks with more than two Cores in total, you will get a bonus! This is called a Combo.\n\nTo continue, get a 4 Combo.\nThis can be easily done using two blocks that have two Cores each, or by clearing three blocks at once with at least one having two Cores. You may need to be patient for the double-Core block you want.";
				// actually, skup this one.
				g->progress++;
				goto setchaintutorial;
			}
		}else if (g->progress==3){
			if (g->p.board->score.maxcombo>=4){
			setchaintutorial:
				newtext="If clearing a block makes the ones above it fall and cause another match, you will earn bonus points. This is called a Chain. The longer the Chain, the more points you will earn.\n\nTo continue, get a x2 chain.";
				newimg="chain.png";
				g->p.board->score.maxchain=0;
			}
		}else if (g->progress==4){
			if (g->p.board->score.maxchain>=2){
				newtext="In regular modes, new blocks will rise from the bottom. You die when the stack tries to rise, but your top row is blocked.\nTo manually raise the stack, press L.\n\nAnd that's it! The controls can be changed on the Main Menu.\nTo exit the tutorial, kill yourself by holding L.";
				g->p.board->score.maxchain=0;
				g->p.board->score.maxcombo=0;
			}
		}else if (g->progress==5){
			if (g->p.board->score.maxchain>=5){
				newtext="Hahaha, I must be a great teacher!\n\nCan you get a 10 chain, I wonder?";
				g->progress++;
			}
			if (g->p.board->score.maxcombo>=8){
				newtext="Consider me mildly impressed.";
			}
		}else if (g->progress==7){
			if (g->p.board->score.maxchain>=10){
				newtext="Okay, don't get so smug. This is just a practice mode.\n\nBut I guess you deserve something. When making a Custom Mode, you can reduce the minimum block length to 1 if you hold L. Pondering why they look ugly is left as an exercise for the reader.";
			}
		}
		if (newtext){
			playSound(winsound,0);
			_modetutorial_queuemessage(g, fTime, newtext);
			g->progress++;
			g->nextimagefname=newimg;
		}
	}
	return ret;
}
static void modetutorial_freeplayable(struct playablegame* g){
	freeSound(((struct playablegame_tutorial*)g)->putse);
}
static void playable_settutorialmode(struct playablegame_tutorial* game, struct gamesettings* settings, struct pcontrols* controls){
	playable_setgeneric(&game->p, settings, controls, NULL);
	game->p.drawfunc=modetutorial_draw;
	game->p.updatefunc=modetutorial_update;
	game->p.initialfunc=modetutorial_initial;
	game->p.newstate=modetutorial_newstate;
	game->p.freestate=modetutorial_freestate;
	game->introdone=0;
	game->putse=loadputstampse();
	game->p.freeplayable=modetutorial_freeplayable;
}
////////////////////////////////////////////////////////////
// MODE - CUSTOM GAME
////////////////////////////////////////////////////////////
struct playablegame_customgame{
	struct playablegame_labels p;
	char* name;
};
static void modecustomgame_freeplayable(struct playablegame* p){
	modelabels_freeplayable(p);
	free(((struct playablegame_customgame*)p)->name);
}
static void modecustomgame_draw(struct gamestate* _g){
	modelabels_draw(_g);
	struct playableextra_labels* lab = &((struct playablegame_labels*)_g->game)->labels;
	if (lab->disabled){
		return;
	}
	struct gamestate_generic* g=(void*)_g;
	const char* custommodename=((struct playablegame_customgame*)g->p.game)->name;
	int customnamew=textWidth(mainfont,custommodename);
	char isoutro=g->boardanimend && g->boardanimdir && !g->retryafteroutro;
	int lx,ly;
	_getlabelpos(lab, g->board, g->p.game->settings, NULL, NULL, NULL, NULL, &lx, &ly, NULL, NULL);
	ly=ly+ezcenter(mainFontHeight,getScreenHeight()-ly);
	lx=lx+ezcenter(customnamew,getScreenWidth()-lx);
	if (isoutro){
		ly+=partMoveFills(g->p.fTime, g->boardanimend, GAMESTATE_GENERIC_OUTROANIMLEN,getScreenHeight()-ly);
	}

	int rectpad=blockw/5;
	drawborderrect(lx-rectpad,ly-rectpad,customnamew+rectpad*2,mainFontHeight+rectpad*2,blockw*.1,0,129,27,0,236,50,255);
	int dropshadowoff=mainFontHeight/10;
	gbDrawText(mainfont, lx+dropshadowoff, ly+dropshadowoff, custommodename, 0,0,0);
	gbDrawText(mainfont, lx, ly, custommodename, 255,255,255);
}
static void playable_setcustomgamemode(struct playablegame_customgame* game, struct gamesettings* settings, struct pcontrols* controls, const char* modename){
	if (settings->isracemode){
		char chosenlabels[] = {4,
							   MODELABEL_TIME,
							   MODELABEL_HIGHTIME,
							   MODELABEL_SCORE,
							   MODELABEL_CURSPEED,
		};
		playable_seteasylabelsmode(&game->p, settings, controls, chosenlabels, NULL);
	}else{
		char chosenlabels[] = {7,
							   MODELABEL_TIME,
							   MODELABEL_HIGHTIME,
							   MODELABEL_SCORE,
							   MODELABEL_HIGHSCORE,
							   MODELABEL_CURSPEED,
							   MODELABEL_MAXCHAIN,
							   MODELABEL_MAXCOMBO,
		};
		playable_seteasylabelsmode(&game->p, settings, controls, chosenlabels, NULL);
	}
	game->name=strdup(modename);
	struct playablegame* asraw=(void*)game;
	asraw->freeplayable=modecustomgame_freeplayable;
	asraw->drawfunc=modecustomgame_draw;
	// TODO - animation cusdtom game name intro
}
////////////////////////////////////////////////////////////
// MODE - FINAL BOSS
////////////////////////////////////////////////////////////
#define FINALBOSSDEATHANIMTIME 5000
#define FINALBOSS_HEALTHBARINPAD (blockw/5)
#define FINALBOSSBGCOLOR 41,0,47
struct playablegame_finalboss{
	struct playablegame p;
	crossTexture* evilface;

	crossTexture* oldbackground;
	crossSE* oldclearsound;
	crossSE* deadlyhit;
	char introdone;
};
struct gamestate_finalboss{
	struct gamestate_generic p;
	int health;
	int maxhealth;
	u64 minchainpopuptime;
	u64 bossdiedat;
	u64 realtime;
	char playedwinsound;
};
static void finalbossdrawbg(unsigned char a){
	drawRectangle(0,0,getScreenWidth(),getScreenHeight(),FINALBOSSBGCOLOR,a);
}
static void _modefinalboss_drawhealth(int x, int y, int w, int h, int cur, int max, unsigned char a){
	drawRectangle(x,y,w,h,255,255,255,a);
	int redw=(w-FINALBOSS_HEALTHBARINPAD*2)*(cur/(double)max);
	if (redw>0){
		drawRectangle(x+FINALBOSS_HEALTHBARINPAD,y+FINALBOSS_HEALTHBARINPAD,redw,h-FINALBOSS_HEALTHBARINPAD*2,255,0,0,a);
	}
}
static void _modefinalboss_calcpos(struct gamestate_finalboss* g, int* healthbarx, int* bary, int* healthbarw, int* barh, int* bossx, int* bossy, int* bossw, int* bossh, unsigned char* a){
	int bx,by,bw,bh;
	stdboarddraw(g->p.board,g->p.p.game->settings,&bx,&by,&bw,&bh);
	int baredgepad=blockw/2;
	*bary=blockw;
	*healthbarx=bx+bw+baredgepad;
	*healthbarw=getScreenWidth()-baredgepad-*healthbarx;
	*barh=getScreenHeight()*.08;
	
	*bossx=*healthbarx;
	*bossw=*healthbarw;
	crossTexture* tex=((struct playablegame_finalboss*)(g->p.p.game))->evilface;
	*bossh=getscaledh(*bossw, tex);
	*bossy=*bary+*barh+baredgepad+ezcenter(*bossh,getScreenHeight()-baredgepad-*bary-*barh);
	*a=255;
	if (g->bossdiedat){
		double dieprogress=(g->realtime-g->bossdiedat)/(double)tohdtime(FINALBOSSDEATHANIMTIME);
		int neww=(1-dieprogress)*(*bossw);
		*bossx+=ezcenter(neww,*bossw);
		*bossw=neww;
		*a=(1-dieprogress)*255;
		int newh=getscaledh(*bossw, tex);
		*bossy+=((getScreenHeight()+newh*dieprogress)-*bossy)*dieprogress;
		*bossh=newh;
	}
}
#define FINALBOSSEXITFADETOBLACKTIME 1000
#define FINALBOSSENTERFADETOBLACKTIME 1000
#define FINALBOSSMANFADEINTIME 2544 // should be same as length of evil laugh
static void modefinalboss_initial(struct gamestate* _g){
	struct playablegame_finalboss* p = ((struct playablegame_finalboss*)(_g->game));
	if (!p->introdone){
		p->introdone=1;
		crossSE* laugh=loadvpkse("assets/se/boss/laugh-evil-1_0.ogg");
		u64 purplefadeinend=getHDTime()+tohdtime(FINALBOSSENTERFADETOBLACKTIME);
		u64 fTime;
		while((fTime=getHDTime())<purplefadeinend){
			startDrawing();
			drawbackground();
			drawRectangle(0,0,getScreenWidth(),getScreenHeight(),FINALBOSSBGCOLOR,partMoveFills(fTime,purplefadeinend,FINALBOSSENTERFADETOBLACKTIME,255));
			endDrawing();
			controlsStart();
			if (menuselectpressed()){
				purplefadeinend=fTime;
			}
			controlsEnd();
		}
		backgroundimg=NULL;
		playSound(laugh,0);
		u64 bossfadeinend=getHDTime()+tohdtime(FINALBOSSMANFADEINTIME);
		while((fTime=getHDTime())<bossfadeinend){
			int barx,bary,barw,barh;
			int bossx,bossy,bossw,bossh;
			unsigned char a;
			_modefinalboss_calcpos((void*)_g, &barx, &bary, &barw, &barh, &bossx, &bossy, &bossw, &bossh, &a);
			startDrawing();
			finalbossdrawbg(255);
			unsigned char introa=partMoveFills(fTime,bossfadeinend,FINALBOSSMANFADEINTIME,255);
			_modefinalboss_drawhealth(barx,bary,partMoveFills(fTime,bossfadeinend,FINALBOSSMANFADEINTIME,barw),barh,1,1,introa);
			crossTexture* tex=p->evilface;
			drawTextureSizedAlpha(tex,bossx,bossy,bossw,bossh,introa);
			endDrawing();
			controlsStart();
			if (isgod && menuselectpressed()){
				bossfadeinend=fTime-1;
			}
			controlsEnd();
		}
		freeSound(laugh);
	}
	modegeneric_initial(_g);
}
static void modefinalboss_freestate(struct gamestate* _g){
	modegeneric_freestate(_g);
}
static struct gamestate* modefinalboss_newstate(struct playablegame* g){
	struct gamestate_finalboss* ret = malloc(sizeof(struct gamestate_finalboss));
	_modegeneric_newstate(g, (void*)ret);
	ret->maxhealth=80;
	ret->health=ret->maxhealth;
	ret->minchainpopuptime=1;
	ret->bossdiedat=0;
	ret->playedwinsound=0;
	((struct gamestate_generic*)ret)->deathaction=(isgod ? 0 : 4);
	return (void*)ret;
}
static void modefinalboss_draw(struct gamestate* _g){
	struct gamestate_finalboss* g = (void*)_g;
	finalbossdrawbg(255);
	modegeneric_draw(_g);
	int barx,bary,barw,barh;
	int bossx,bossy,bossw,bossh;
	unsigned char a;
	_modefinalboss_calcpos((void*)_g, &barx, &bary, &barw, &barh, &bossx, &bossy, &bossw, &bossh, &a);
	if (!g->bossdiedat){
		_modefinalboss_drawhealth(barx,bary,barw,barh,g->health,g->maxhealth, 255);
	}else{
		if (g->playedwinsound){
			u64 fadeoutendtime=g->bossdiedat+tohdtime(FINALBOSSDEATHANIMTIME+FINALBOSSEXITFADETOBLACKTIME);
			int recta=partMoveFills(g->realtime, fadeoutendtime, FINALBOSSEXITFADETOBLACKTIME, 255);
			//drawRectangle(0,0,getScreenWidth(),getScreenHeight(),0,0,0,recta);
			drawTextureSizedAlpha(((struct playablegame_finalboss*)_g->game)->oldbackground,0,0,getScreenWidth(),getScreenHeight(),recta);
			a=0;
		}
	}
	drawTextureSizedAlpha(((struct playablegame_finalboss*)(g->p.p.game))->evilface,bossx,bossy,bossw,bossh,a);
	if (g->p.boardanimend && g->p.boardanimdir && !g->p.retryafteroutro){
		drawRectangle(0,0,getScreenWidth(),getScreenHeight(),0,0,0,partMoveFills(g->p.p.fTime, g->p.boardanimend, GAMESTATE_GENERIC_OUTROANIMLEN,255));
	}
}
static int modefinalboss_update(struct gamestate* _g){
	struct gamestate_finalboss* g = (void*)_g;
	int ret;
	if (g->bossdiedat){
		ret=0;
		g->realtime=_g->fTime-g->p.timeOff;
		_g->fTime=g->bossdiedat;
		controlsReset();
	}else{
		ret=modegeneric_update(_g);
		g->realtime=_g->fTime;
	}
	struct gboard* b = ((struct gamestate_generic*)_g)->board;
	if (g->bossdiedat){
		if (g->realtime>g->bossdiedat+tohdtime(FINALBOSSDEATHANIMTIME)){
			if (g->realtime>g->bossdiedat+tohdtime(FINALBOSSDEATHANIMTIME+FINALBOSSEXITFADETOBLACKTIME)){
				return 2;
			}else if (!g->playedwinsound){
				g->playedwinsound=1;
				playSound(winsound,0);
			}
		}
	}else if (!b->diedat){
		u64 newminpopuptime=0;
		int totaldamage=0;
		// horrible way of detecting chains by reading popups
		ITERATENLIST(b->popups,{
				struct chainpopup* c = _curnList->data;
				if (c->timeend>g->minchainpopuptime){
					if (c->timeend>newminpopuptime){
						newminpopuptime=c->timeend;
					}
					if (c->ischain){
						totaldamage+=c->value;
					}else{
						totaldamage+=c->value-2;
					}
				}
			});
		if (newminpopuptime){
			g->minchainpopuptime=newminpopuptime;
		}
		if (totaldamage){
			g->health-=totaldamage;
			if (g->health<=0){
				g->health=0;
				/* g->p.board->diedat=_g->fTime; */
				/* g->p.deathaction=1; */
				_g->game->settings->doautoupshift=0;
				g->p.board->queuedUpshiftTime=0;
				g->bossdiedat=_g->fTime;

				struct playablegame_finalboss* p = (void*)g->p.p.game;
				playSound(p->deadlyhit,0);
				stopMusic(p->p.musichandle);
				freeMusic(p->p.music);
				p->p.music=NULL;

				if (!isgod){
					isgod=1;
					char* trophypath=getisgodpath();
					FILE* fp = fopen(trophypath,"wb");
					if (fp){
						time_t gotdate=time(NULL);
						fwrite(&(gotdate),sizeof(gotdate),1,fp);
						fclose(fp);
					}
					free(trophypath);
				}
			}
		}
	}
	return ret;
}
static void modefinalboss_freeplayable(struct playablegame* p){
	struct playablegame_finalboss* game=(void*)p;
	backgroundimg=game->oldbackground;
	freeTexture(game->evilface);
	
	freeSound(gsoundsloaded[GSOUND_STARCONNECT]);
	gsoundsloaded[GSOUND_STARCONNECT]=game->oldclearsound;
	freeSound(game->deadlyhit);
}
static void playable_setfinalboss(struct playablegame_finalboss* game, struct gamesettings* settings, struct pcontrols* controls){
	crossMusic* m;
	m=loadvpkmusic("assets/se/puzzleitnes_0.ogg");
	setMusicVolumeBefore(m,128*.7);
	
	playable_setgeneric(&game->p, settings, controls, m);
	game->p.drawfunc=modefinalboss_draw;
	game->p.updatefunc=modefinalboss_update;
	game->p.initialfunc=modefinalboss_initial;
	game->p.newstate=modefinalboss_newstate;
	game->p.freestate=modefinalboss_freestate;
	game->p.freeplayable=modefinalboss_freeplayable;

	game->oldbackground=backgroundimg;
	game->evilface=loadvpkpng("assets/img/evil.png");
	game->oldclearsound=gsoundsloaded[GSOUND_STARCONNECT];
	gsoundsloaded[GSOUND_STARCONNECT]=loadvpkse("assets/se/boss/normal-attack.ogg");
	game->deadlyhit=loadvpkse("assets/se/boss/deadlyhit.ogg");
	game->introdone=0;
	
}
////////////////////////////////////////////////////////////
static int isfnameinvalidchar(char c){
	static char forbiddenchars[]={'"','*','/',':','<','>','?','\\','|'};
	for (int i=0;i<sizeof(forbiddenchars);++i){
		if (c==forbiddenchars[i]){
			return 1;
		}
	}
	if (c<=0x1f){
		return 1;
	}
	return 0;
}
static char* selectcustommode(int ismakemode){
	int numfiles;
	char** _list = getfilelist(custommodedir, 2, &numfiles);
	if (ismakemode){
		memmove(_list+2,_list,sizeof(char*)*numfiles);
		_list[0]=strdup("<Back>");
		_list[1]=strdup("<New>");
		numfiles+=2;
	}else{
		if (numfiles==0){
			_list[0]=strdup("No modes made.");
		}else{
			_list[numfiles]=strdup("<Back>");
		}
		numfiles++;
	}
	int choice=simplecmenu(_list, numfiles, 0,NULL);
	char* ret;
	if (choice==-1 || (ismakemode && choice==0) || (!ismakemode && choice==numfiles-1)){
		ret=NULL;
	}else{
		if (ismakemode && choice==1){
			ret=malloc(1);
			ret[0]='\0';
		}else{
			ret=easyCombineStrings(2,custommodedir,_list[choice]);
		}
	}
	for (int i=0;i<numfiles;++i){
		free(_list[i]);
	}
	free(_list);
	return ret;
}
// returns 1 if all the data was set up
static int custommodemenu(struct gamesettings** se, char** modename){
top:;
	char* chosenpath;
	char ismakemode;
	{
		char* labels[] = {"Play Custom Mode","Make Custom Mode"};
		static int lastchoice=0;
		int choice=simplecmenu(labels, sizeof(labels)/sizeof(char*), lastchoice,NULL);
		if (choice==-1){
			return 0;
		}
		lastchoice=choice;
		chosenpath=selectcustommode(choice==1);
		ismakemode=(choice==1);
	}
	if (!chosenpath){
		goto top;
	}
	if (ismakemode){
		char* defaultfname;
		struct gamesettings editme;
		if (chosenpath[0]=='\0'){ // special signal. means to make a new mode.
			defaultfname=strdup("mymode");
			initgamesettings(&editme);
		}else{
			if (loadsettings(chosenpath, &editme)){
				sadflash();
				free(chosenpath);
				goto top;
			}
			defaultfname=dupjustfname(chosenpath);
		}
		free(chosenpath);
		chosenpath=NULL;
	reeditsettings:
		if (!ruleseteditor(&editme)){
			free(defaultfname);
			goto top;
		}
	reentername:;
		char* newfname=userKeyboardInput(defaultfname, "Enter filename to save mode", 50);
		if (!newfname){
			goto reeditsettings;
		}
		for (int i=0;i<strlen(newfname);++i){
			if (isfnameinvalidchar(newfname[i])){
				newfname[i]=' ';
			}
		}
		char* writepath=easyCombineStrings(2,custommodedir,newfname);
		free(newfname);
		if (checkFileExist(writepath)){
			char* labels[]={"Overwrite","Back"};
			if (simplecmenu(labels, 2, 0,NULL)!=0){
				free(writepath);
				goto reentername;
			}
		}

		// It's finally time
		FILE* fp = fopen(writepath,"wb");
		if (!fp){
			// TODO - show warning
			free(writepath);
			sadflash();
			goto reentername;
		}
		free(defaultfname);
		writesettings(&editme,fp);
		fclose(fp);

		// now play the mode we made.
		*se=malloc(sizeof(struct gamesettings));
		*(*se)=editme;
		*modename=dupjustfname(writepath);
		free(writepath);
		return 1;
	}else{
		struct gamesettings loadedsettings;
		if (loadsettings(chosenpath, &loadedsettings)){
			sadflash();
			free(chosenpath);
			return 0;
		}
		*se=malloc(sizeof(struct gamesettings));
		*(*se)=loadedsettings;
		*modename=dupjustfname(chosenpath);
		free(chosenpath);
		return 1;
	}
	return 0;
}

#define TITLESCREENANIMEND 1000
#define TITLESCREENPARTICLES 50
#define TITLESCREENHANDTIME 100
#define TITLESCREENPROMPTFLASHRATE 1000
static void _titlescreen(crossMusic* menumusic, crossPlayHandle* rethandle){
	crossTexture* logo = loadvpkpng("assets/img/ui/logo.png");
	int logow=getTextureWidth(logo);
	int logoh=getTextureHeight(logo);
	char logoanimdone=0;
	char isoutro=0;
	u64 logoSpinEnd=getHDTime()+tohdtime(TITLESCREENANIMEND);
	u64 handendtime=0;
	crossTexture* hand=NULL;
	u64 nextpressstarttoggle=0;
	char showpressstart=0;
	
	// i really need to rush to get this game done
	int particleX[TITLESCREENPARTICLES];
	int particleY[TITLESCREENPARTICLES];
	mltime particleDTime[TITLESCREENPARTICLES];
	u64 particleEndTime[TITLESCREENPARTICLES];
	double particlescales[TITLESCREENPARTICLES];
	double particlestartrot[TITLESCREENPARTICLES];
	double particleendrot[TITLESCREENPARTICLES];
	char particleimages[TITLESCREENPARTICLES];
	char dooffscreenparticles=1;
	
	while(1){
		u64 fTime=getHDTime();
		if (isoutro && fTime>=logoSpinEnd){
			break;
		}
		if (fTime>=nextpressstarttoggle && nextpressstarttoggle){
			showpressstart=!showpressstart;
			nextpressstarttoggle=fTime+tohdtime(TITLESCREENPROMPTFLASHRATE);
		}
		startDrawing();
		drawbackground();
		double logoscale=1;
		double logoangle=0;
		if (fTime>=logoSpinEnd){
			if (!logoanimdone){
				logoanimdone=1;
				if (menumusic){
					*rethandle=playMusic(menumusic,0);
				}
				for (int i=0;i<TITLESCREENPARTICLES;++i){
					particleEndTime[i]=fTime;
				}
				nextpressstarttoggle=1;
			}
		}else{
			logoscale=partMoveFillsR(fTime, logoSpinEnd, TITLESCREENANIMEND, 1, isoutro);
			logoangle=partMoveFillsR(fTime, logoSpinEnd, TITLESCREENANIMEND, M_PI*4, isoutro);
		}
		unsigned char versionnumbera=255;
		unsigned char a=255;
		if (fTime<logoSpinEnd){
			versionnumbera=partMoveFillsR(fTime,logoSpinEnd,TITLESCREENANIMEND,255,isoutro);
			if (isoutro){
				a=versionnumbera;
			}
		}
		gbDrawTextAlpha(mainfont,0,getScreenHeight()-mainFontHeight,VERSIONSTRING,255,255,255,versionnumbera);
		
		if (logoanimdone){
			for (int i=0;i<TITLESCREENPARTICLES;++i){
				if (fTime>=particleEndTime[i]){
					// gen new particle
					particlescales[i]=randbetween(1.3,3);
					particleimages[i]=randbetween(0,maxcolors-1);
					crossTexture* tex=drawcolors[particleimages[i]].heart;
					particleX[i]=randbetween(getTextureWidth(tex)*particlescales[i]*-1,getScreenWidth()-1);
					particleDTime[i]=randbetween(500,4000);
					particleY[i]=getTextureHeight(tex)*particlescales[i]*-1;
					if (dooffscreenparticles){
						int extraYOff=randbetween(0,getScreenHeight()*1.5);
						particleY[i]-=extraYOff;
						particleDTime[i]*=extraYOff/getScreenHeight();
					}
					particleEndTime[i]=fTime+tohdtime(particleDTime[i]);
					particlestartrot[i]=randbetween(0,M_PI*2);
					particleendrot[i]=randbetween(0,M_PI*10);
				}else{
					double progress=partMoveFills(fTime, particleEndTime[i], particleDTime[i], 1);
					int y=particleY[i]+(getScreenHeight()-particleY[i])*progress;
					double angle=particlestartrot[i]+particleendrot[i]*progress;
					crossTexture* tex=drawcolors[particleimages[i]].heart;
					double scale=particlescales[i];
					drawTextureSizedAlphaRotate(tex, particleX[i], y, getTextureWidth(tex)*scale, getTextureHeight(tex)*scale, a, angle);
				}
			}
			dooffscreenparticles=0;
		}
		int destx=ezcenter(logow*logoscale,getScreenWidth());
		int desty=ezcenter(logoh*logoscale,getScreenHeight()*.3);
		drawTexturePartSizedTintAlphaRotate(logo, destx, desty, logow*logoscale, logoh*logoscale, 0,0,logow,logoh,255,255,255,a, logoangle, 0);
		if (handendtime){
			int w=getTextureWidth(hand);
			int h=getTextureHeight(hand);
			int x=getScreenWidth()-w;
			int y=getScreenHeight()-h;
			if (fTime<handendtime){
				x+=partMoveEmpties(fTime, handendtime, TITLESCREENHANDTIME, w);
				y+=partMoveEmpties(fTime, handendtime, TITLESCREENHANDTIME, h);
			}
			drawTextureAlpha(hand,x,y,a);
		}
		if (showpressstart && !isoutro){
			gbDrawTextfCenter(mainfont, 0, getScreenHeight()*.6+ezcenter(mainFontHeight,getScreenHeight()*.3), 0,0,0, 255, getScreenWidth(), "PRESS START");
		}
		
		endDrawing();
		controlsStart();
		if (menuselectpressed()){
			playSound(menuselectsound,0);
			if (!logoanimdone){
				logoSpinEnd=fTime-1;
			}else{
				if (isDown(BUTTON_UP) && isDown(BUTTON_LEFT) && isDown(BUTTON_L) && !hand){
					hand=loadvpkpng("assets/img/ui/gamemodelogo/egg.png");
					handendtime=fTime+tohdtime(TITLESCREENHANDTIME);
				}else{
					if (isDown(BUTTON_L) && isDown(BUTTON_R)){
						blockw=ceil(((double)blockw)*.66);
					}
					if (!isoutro){
						// outro animation
						isoutro=1;
						logoSpinEnd=getHDTime()+tohdtime(TITLESCREENANIMEND);
					}else{
						logoSpinEnd=fTime-1;
					}
				}
			}
		}
		controlsEnd();
	}
	if (hand){
		freeTexture(hand);
	}
	freeTexture(logo);
}
static void _mainmenudrawtrophies(struct cmenu* m, u64 fTime){
	if (isgod && (!m->isoutro || m->enterAnimationEnd)){
		int insidepad=blockw/4;
		int labelboxw=m->maxlabelw+insidepad*2;
		int labelx=ezcenter(labelboxw,getScreenWidth());
		double scale=1;
		if (m->enterAnimationEnd){
			scale=partMoveFillsR(fTime, m->enterAnimationEnd, CMENUENTERANIMTIME, 1, m->isoutro);
		}
		int myheartw=blockw*2.5*scale;
		int myheartx=labelx+labelboxw;
		myheartx+=ezcenter(myheartw,getScreenWidth()-myheartx);
		int myhearty=ezcenter(myheartw,getScreenHeight());
		drawTextureSized(drawcolors[1].heart,myheartx,myhearty,myheartw,myheartw);
	}
}
static const char standardmodelabels[] = {6,
									MODELABEL_TIME,
									MODELABEL_SCORE,
									MODELABEL_HIGHSCORE,
									MODELABEL_CURSPEED,
									MODELABEL_MAXCHAIN,
									MODELABEL_MAXCOMBO};
enum mainmenuoptions{
	MAINMENU_NORMALMODE=0,
	MAINMENU_EASYMODE=1,
	MAINMENU_TUTORIAL=2,
	MAINMENU_PRACTICE=3,
	MAINMENU_QUADCOLORRACE=4,
	MAINMENU_LOWIQRACE=5,
	MAINMENU_DEATHMODE=6,
	MAINMENU_FINALBOSS=7,
	MAINMENU_CUSTOMGAME=8,
	MAINMENU_CONTROLS=9,
	
	MAINMENU_MAX,
};
static int mainmenu(char doshowtitle){
verytop:;
	crossMusic* titlebgm=loadvpkmusic("assets/se/BobsAdventures-back34_0.ogg");
	setMusicVolumeBefore(titlebgm,127);
	crossPlayHandle playhandle;
	if (doshowtitle){
		_titlescreen(titlebgm,&playhandle);
		doshowtitle=0;
	}else{
		playhandle=playMusic(titlebgm,0);
	}
backtomenu:;
	struct playablegame* createdgame=NULL;
	char* labels[MAINMENU_MAX];
	labels[MAINMENU_NORMALMODE]="Normal Mode";
	labels[MAINMENU_EASYMODE]="Easy Mode";
	labels[MAINMENU_TUTORIAL]="Tutorial";
	labels[MAINMENU_PRACTICE]="Practice";
	labels[MAINMENU_QUADCOLORRACE]="4 Color Race";
	labels[MAINMENU_LOWIQRACE]="Low IQ Race";
	labels[MAINMENU_DEATHMODE]="Death Mode";
	labels[MAINMENU_FINALBOSS]="Final Boss";
	labels[MAINMENU_CUSTOMGAME]="Custom Game";
	labels[MAINMENU_CONTROLS]="Controls";
	int numoptions=MAINMENU_MAX;
	static int lastchoice=0;
	int choice=simplecmenu(labels, numoptions, lastchoice,_mainmenudrawtrophies);
	if (choice!=-1){
		lastchoice=choice;
	}
	if (choice==MAINMENU_CUSTOMGAME){
		struct gamesettings* se;
		char* modename;
		if (!custommodemenu(&se,&modename)){
			goto backtomenu;
		}
		createdgame=malloc(sizeof(struct playablegame_customgame));
		char* highscorepath=easyCombineStrings(2,customgamescorefilesdir,modename);
		playable_setcustomgamemode((void*)createdgame, se, &globalcontrols, modename);
		free(modename);
		createdgame->highscoredatapath=highscorepath;
		goto actuallyitdoeschaintogame;
	}else if (choice==MAINMENU_CONTROLS){
		controlseditor(&globalcontrols);
		char* controlspath=getsavedcontrolspath();
		savecontrols(controlspath, &globalcontrols, preferreddasdelay, preferreddasrepeat);
		free(controlspath);
		goto backtomenu;
	}else if (choice==-1){
		_titlescreen(NULL,NULL);
		goto backtomenu;
	}else{ // options that don't lead to a submenu
	actuallyitdoeschaintogame:;
		u64 endfadeouttime=getHDTime()+tohdtime(300);
		fadeoutMusic(playhandle,300);
		if (choice==MAINMENU_EASYMODE){ // easy
			createdgame=malloc(sizeof(struct playablegame_labels));
			struct gamesettings* se = malloc(sizeof(struct gamesettings));
			initgamesettings(se);
			se->falltime+=30;
			se->poptime+=200;
			se->timetodie+=1000;
			se->initialupshifttime=14000;
			se->delayshortentime*=4;
		
			playable_seteasylabelsmode((void*)createdgame, se, &globalcontrols, standardmodelabels, "assets/img/ui/gamemodelogo/moderngames.png");
			createdgame->highscoredatapath=easyCombineStrings(2,scorefilesdir,"easymode");
		}else if (choice==MAINMENU_NORMALMODE){ // normal
			createdgame=malloc(sizeof(struct playablegame_labels));
			struct gamesettings* se = malloc(sizeof(struct gamesettings));
			initgamesettings(se);
			playable_seteasylabelsmode((void*)createdgame, se, &globalcontrols, standardmodelabels, "assets/img/ui/gamemodelogo/normalmode.png");
			createdgame->highscoredatapath=easyCombineStrings(2,scorefilesdir,"normalmode");
		}else if (choice==MAINMENU_PRACTICE){ // practice
			createdgame=malloc(sizeof(struct playablegame_labels));
			struct gamesettings* se = malloc(sizeof(struct gamesettings));
			initgamesettings(se);
			se->doautoupshift=0;
			char chosenlabels[] = {6,
								   MODELABEL_TIME,
								   MODELABEL_SCORE,
								   MODELABEL_MAXCHAIN,
								   MODELABEL_BESTCHAIN,
								   MODELABEL_MAXCOMBO,
								   MODELABEL_BESTCOMBO,
								   MODELABEL_BLOCKSCLEARED,
			};
			playable_seteasylabelsmode((void*)createdgame, se, &globalcontrols, chosenlabels, "assets/img/ui/gamemodelogo/practice.png");
			createdgame->highscoredatapath=easyCombineStrings(2,scorefilesdir,"practice");
		}else if (choice==MAINMENU_FINALBOSS){ // final boss
			createdgame=malloc(sizeof(struct playablegame_labels));
			struct gamesettings* se = malloc(sizeof(struct gamesettings));
			initgamesettings(se);
			se->gengarbage=1;
			se->initialblockgen=0;
			se->playerplacetime/=2;
			se->delayshortenratio=0;
			se->initialupshifttime=2000;
			se->isracemode=1;
			playable_setfinalboss((void*)createdgame, se, &globalcontrols);
			createdgame->highscoredatapath=easyCombineStrings(2,scorefilesdir,"finalbossunused");
		}else if (choice==MAINMENU_DEATHMODE){ // death
			createdgame=malloc(sizeof(struct playablegame_labels));
			struct gamesettings* se = malloc(sizeof(struct gamesettings));
			initgamesettings(se);
			se->initialblockgen=5;
			se->initialupshifttime=1300;
			se->delayshortenratio=.1;
			playable_setdeathmode((void*)createdgame, se, &globalcontrols);
			createdgame->highscoredatapath=easyCombineStrings(2,scorefilesdir,"death");
		}else if (choice==MAINMENU_LOWIQRACE){ // low iq race
			createdgame=malloc(sizeof(struct playablegame_labels));
			struct gamesettings* se = malloc(sizeof(struct gamesettings));
			initgamesettings(se);
			se->initialblockgen=150;
			se->defaultboardw=18;
			se->numcolors=1;
			se->doautoupshift=0;
			se->isracemode=1;
			se->canmanualupshift=0;
			playable_setracemode((void*)createdgame, se, &globalcontrols, "assets/img/ui/gamemodelogo/black.png"); 
			createdgame->highscoredatapath=easyCombineStrings(2,scorefilesdir,"lowiqrace");
		}else if (choice==MAINMENU_QUADCOLORRACE){ // 4 color race
			createdgame=malloc(sizeof(struct playablegame_labels));
			struct gamesettings* se = malloc(sizeof(struct gamesettings));
			initgamesettings(se);
			se->initialblockgen=150;
			se->defaultboardw=18;
			se->numcolors=4;
			se->doautoupshift=0;
			se->canmanualupshift=0;
			se->isracemode=1;
			playable_setracemode((void*)createdgame, se, &globalcontrols, "assets/img/ui/gamemodelogo/4colorrace.png");
			createdgame->highscoredatapath=easyCombineStrings(2,scorefilesdir,"4colorrace");
		}else if (choice==MAINMENU_TUTORIAL){ // tutorial
			createdgame=malloc(sizeof(struct playablegame_tutorial));
			struct gamesettings* se = malloc(sizeof(struct gamesettings));
			initgamesettings(se);
			se->doautoupshift=0;
			se->initialblockgen=0;
			playable_settutorialmode((void*)createdgame, se, &globalcontrols);
			createdgame->highscoredatapath=easyCombineStrings(2,scorefilesdir,"tutorialscoresecret");
		}
		// wait for bgm fadeout
		while(getHDTime()<endfadeouttime){
			controlsStart();
			if (menuselectpressed()){
				break;
			}
			controlsEnd();
			wait(1);
		}
		stopMusic(playhandle);
	}
	freeMusic(titlebgm);
	///
	if (createdgame){
		int playgameret=playgame(createdgame);
		free(createdgame->highscoredatapath);
		free(createdgame->settings);
		free(createdgame);
		if (playgameret==3){
			doshowtitle=1;
		}
	}
	goto verytop;
	return 0; // suppress warning
}
int main(int argc, char** argv){
	init();
	mainmenu(1);
	quitGraphics();
}
